--[[----------------------------------------------------------------------------

LrmCacheIndexdefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local KSDate = require "KSDate"

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--

-- @mark LrmCacheIndexdefinitions namespace

--------------------------------------------------------------------------------

local LrmCacheIndexdefinitions = {}

--------------------------------------------------------------------------------

--[[
	These constants are only used when existing photos (imported on 2011-08-12
	and before are in the system and we didn't capture the rendition size.
]]--

local defaultSizeForRendition = {
	original = 20000000,
	proxy = 943718,
	preview = 150000,
	thumbnail = 12000,
	thumbnail2x = 48000,
	xmp_develop = 16384
}

--------------------------------------------------------------------------------

function LrmCacheIndexdefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	-- table cacheReferences

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE cacheReferences( path, renditionName, contentId, size, userUpdated, hasUploaded )]],

		updateLazily = false,
		
		docValueFilter = function( docValue )
			return (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset'
		end,

		map = function( docValue, emit )

			if docValue.type ~= 'asset' then
				return
			end
					
			local serverLinks =
				  ( docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links ) or {}

			local function emitForRenditionName( renditionName, ozNameForRendition )
			
				local renditionInfo = derefOpt( docValue, '_localOnly', 'files', renditionName )
				
				if not renditionInfo then
					return
				end
				
				local path = renditionInfo.path
				if path then
				
					local serverLinksForRendition = serverLinks[ ozNameForRendition ]
					local shouldUpload
					
					if renditionName == 'master' then
						-- we aren't making destructive edits to originals yet, so we do not upload
						-- originals other than the base original
						if derefOpt( docValue, 'importSource', 'sha256' ) ~= renditionInfo.contentId then
							-- We don't upload the latest original
							shouldUpload = false
						else
							shouldUpload = ( not serverLinksForRendition )
								or serverLinksForRendition.invalid
								or renditionInfo.revId == 'pending'
						end
					elseif renditionName ~= 'proxy' then
						shouldUpload = ( not serverLinksForRendition )
							or serverLinksForRendition.invalid
							or renditionInfo.revId == 'pending'
					else
						shouldUpload = false
					end

					if renditionName == 'xmp_develop' then
						-- INFO [immisch 2018-07-20]: delayed migration
						renditionInfo.contentId = renditionInfo.contentId or renditionInfo.sha256
					end

					local size = renditionInfo.fileSize
						or defaultSizeForRendition[ renditionName ] or 1000000

					emit( path,
						  renditionName,
						  renditionInfo.contentId,
						  size,
						  docValue.userUpdated or KSDate.timeInIso8601Format(),
						  shouldUpload and 0 or 1 )

--					cacheReferencesLog:debug( "photoId = %s | renditionName = %s | size = %s | epoch  %s | hasUploaded = %s",
--						docValue._docId, renditionName, renditionInfo.size, renditionInfo.epoch or 1,
--						not shouldUpload )
				
				end
			
			end

			emitForRenditionName( 'thumbnail', '/rels/rendition_type/144h' )
			emitForRenditionName( 'thumbnail2x', '/rels/rendition_type/thumbnail2x' )
			emitForRenditionName( 'preview', '/rels/rendition_type/2048' )
			emitForRenditionName( 'proxy', '/rels/proxy_type/2560' )
			emitForRenditionName( 'master', '/rels/master' )
			emitForRenditionName( 'xmp_develop', '/rels/xmp/develop' )

			local xmpDevelop = derefOpt( docValue, '_localOnly', 'files', 'xmp_develop' )
			if xmpDevelop and xmpDevelop.path then
				emit( xmpDevelop.path,
					 'xmp_develop',
					 xmpDevelop.size or defaultSizeForRendition[ 'xmp_develop' ] or 16434,
					 1,
					 ( xmpDevelop.uploadStatus and xmpDevelop.uploadStatus == "success" ) and 1 or 0 )

			end
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX cacheReferences_path ON cacheReferences( path ASC, docId )]]
	indexStoreSchema:addIndex [[CREATE INDEX cacheReferences_potentialVictims
		ON cacheReferences( userUpdated ASC, size DESC, docId )]]
	indexStoreSchema:addIndex [[CREATE INDEX cacheReferences_hasUploaded ON cacheReferences( hasUploaded )]]
	indexStoreSchema:addIndex [[CREATE INDEX cacheReferences_renditionName ON cacheReferences( renditionName )]]

	-- table avatarCacheReferences
	
	indexStoreSchema:addTable {

		sql = [[CREATE TABLE avatarCacheReferences( md5, avatarType, size, epoch )]],

		docValueFilter = function( docValue )
			return docValue.type == 'feedbackuser'
		end,

		map = function( docValue, emit )
		
			local docType = docValue.type
			
			if docType == 'feedbackuser' then
			
				local avatars = derefOpt( docValue, 'avatars' )
				
				local function emitForAvatarType( avatarType )
				
					local info = derefOpt( avatars, avatarType )
					
					if info then
						emit( info.md5, avatarType, info.size, info.epoch )
					end
					
				end
				
				emitForAvatarType( 'small' )
				emitForAvatarType( 'small2x' )
				emitForAvatarType( 'medium' )
				emitForAvatarType( 'medium2x' )
				emitForAvatarType( 'large' )
				emitForAvatarType( 'large2x' )
				
			end
			
		end
		
	}

end

--------------------------------------------------------------------------------

return LrmCacheIndexdefinitions
