--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'

local KSTableUtils = require "KSTableUtils"

local LriAssetUtils = require 'LriAssetUtils'

local LriCoreInfoIndexDefinitions = require "LriCoreInfoIndexDefinitions"

local LriOzRenditionServerUtils = require "LriOzRenditionServerUtils"

local WFRxIndexStoreSchema = require 'WFRxIndexStoreSchema'

local WFOzContentTypes = require 'WFOzContentTypes'

local WFCapabilities = require 'WFCapabilities'

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

local assetIsDisplayable = assert( LriCoreInfoIndexDefinitions.assetIsDisplayable )

local getContentTypeNumberForContentType = WFOzContentTypes.getContentTypeNumberForContentType

--==============================================================================

local LriCatalogIndexDefinitions = {}

--------------------------------------------------------------------------------
-- @mark assetProfileLinks
--------------------------------------------------------------------------------

local kAssetProfileLinks = {

	sql = [[CREATE TABLE assetProfileLinks( subtype, fileName, masterUrl, uniqueId )]],
	
	docValueFilter = function( docValue )
		return (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset'
	end,

	map = function( docValue, emit )
		
		if docValue._deleted then
			return
		end
		if not assetIsDisplayable( docValue ) then
			return
		end
		
		local cameraProfileInfo = LriAssetUtils.getCameraProfileInfo( docValue)
		if cameraProfileInfo then
			emit( "camera", cameraProfileInfo.filename, cameraProfileInfo.href, cameraProfileInfo.uniqueId )
		end
		
		local lensProfileInfo = LriAssetUtils.getLensProfileInfo( docValue )
		if lensProfileInfo then
			emit( "lens", lensProfileInfo.filename, lensProfileInfo.href, lensProfileInfo.uniqueId )
		end
	end,
}

local kAssetProfileLinks_uniqueId = [[CREATE INDEX assetProfilesLink_uniqueId ON assetProfileLinks( uniqueId )]]

--------------------------------------------------------------------------------
-- @mark profileRegistration
--------------------------------------------------------------------------------

local kProfileRegistration = {

	sql = [[CREATE TABLE profileRegistration( subtype, fileName, uniqueId )]],

	docValueFilter = function( docValue )
		return (docValue.type == 'asset' and docValue.subtype == 'camera_profile') or docValue.type == 'profile'
	end,

	map = function( docValue, emit )

		if (docValue.type == 'asset' and docValue.subtype == 'camera_profile') or docValue.type == 'profile'
			and not docValue._deleted then

			local fileName
			local uniqueId
			local subtype

			if docValue.type == 'asset' then
				-- Custom DCPs

				subtype = 'camera'
				fileName = docValue.fileName
				uniqueId = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )

			else
				-- Adobe standard camera matching profiles

				local localOnly = docValue._localOnly

				subtype = localOnly.subtype
				fileName = localOnly.fileName
				uniqueId = localOnly.uniqueId
			end

			if uniqueId then
				emit( subtype, fileName, uniqueId )
			end
		end
	end,
}

local kProfileRegistration_uniqueId = [[CREATE INDEX profileRegistration_uniqueId ON profileRegistration( uniqueId )]]
local kProfileRegistration_fileName = [[CREATE INDEX profileRegistration_fileName ON profileRegistration( fileName COLLATE NOCASE )]]

--------------------------------------------------------------------------------
-- @mark TEMP_searchFilterCount
--------------------------------------------------------------------------------

local kTEMP_searchFilterCount =  {

	sql = [[CREATE TEMPORARY TABLE TEMP_searchFilterCount( filterType, filterCount )]],
	
	isManualTable = true
	
}

--------------------------------------------------------------------------------
-- @mark renditionRevisions
--------------------------------------------------------------------------------

local function isVideo( docValue )

	if docValue and docValue.video then
		
		return true
		
	end
	
	local contentType = derefOpt( docValue, 'importSource', 'contentType' )
								or derefOpt( docValue, '_serverProvidedMetadata', 'links', 
												'/rels/master', 'content_type' )
				
	if contentType then
	
		local contentTypeNumber = getContentTypeNumberForContentType( contentType )
		
		if LriAssetUtils.isVideoContentTypeNumber( contentTypeNumber ) then
	
			return true
			
		end
		
	end
	
	return false
		
end

--------------------------------------------------------------------------------

local kRenditionRevisionsDefinition = {

	sql = [[CREATE TABLE renditionRevisions( renditionName, timestamp, size )]],
	
	docValueFilter = function( docValue )
		return (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset'
	end,

	map = function( docValue, emit )

		if docValue._deleted then
			return
		end
		
		if not assetIsDisplayable( docValue ) then return end
		
		local timestamp = docValue
			and docValue.importSource
			and docValue.importSource.importTimestamp
		
		if LriOzRenditionServerUtils.shouldDownloadOzRendition( docValue, "thumbnail2x" ) then
		
			local fileSize = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/rendition_type/thumbnail2x', 'fileSize' )
				or 48000
		
			emit( 'thumbnail2x', timestamp, fileSize )
		end
		
		if LriOzRenditionServerUtils.shouldDownloadOzRendition( docValue, "preview" ) then
		
			local fileSize = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/rendition_type/2048', 'fileSize' )
				or 461824
	
			emit( 'preview', timestamp, fileSize )
		end
		
		local hasProxyOnOz = LriAssetUtils.hasProxyOnOz( docValue )
		local hasProxyLocally = LriAssetUtils.hasProxyLocally( docValue )
		local hasMasterOnOz = LriAssetUtils.hasMasterOnOz( docValue )
		local hasMasterLocally = LriAssetUtils.hasMasterLocally( docValue )
		local hasExternalXmpDevelop = LriAssetUtils.hasExternalXmpDevelop( docValue )
		
		local isVideo = isVideo( docValue )
		
		if not isVideo then
		
			if hasMasterOnOz and not hasMasterLocally then
			
				local fileSize = derefOpt( docValue, 'importSource', 'fileSize' )
				if not fileSize or fileSize <= 0 then
					fileSize = 20971520
				end
		
				emit( 'master', timestamp, fileSize )
				
			end
			
			if hasProxyOnOz and not hasProxyLocally then
			
				local fileSize = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/proxy_type/2560', 'fileSize' ) 				
				if not fileSize then 
					fileSize = derefOpt( docValue, '_localOnly', 'proxy', 'fileSize' ) or 1331692
				end
			
				emit( 'proxy', timestamp, fileSize )
			end
			
			if not hasMasterOnOz and not hasMasterLocally and not hasProxyLocally then
				emit( 'proxyWithoutMaster', timestamp, 1331692 )
			end
		
			if hasExternalXmpDevelop and not LriAssetUtils.hasExternalXmpDevelopLocally( docValue ) then
			
				local invalid = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'invalid' )

				if not invalid then

					local fileSize = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'fileSize' )
						or 16384

					emit( 'xmp_develop', timestamp, fileSize )
				end
			end

		end
						
	end
}


function LriCatalogIndexDefinitions.tableDefinitions( tableDefinitions )
	
	tableDefinitions.renditionRevisions = kRenditionRevisionsDefinition
	tableDefinitions.assetProfileLinks = kAssetProfileLinks
	tableDefinitions.profileRegistration = kProfileRegistration
	tableDefinitions.TEMP_searchFilterCount = kTEMP_searchFilterCount
	
	tableDefinitions.new_asset_activity = nil
	tableDefinitions.tags = nil
	tableDefinitions.favorite_activity = nil
	tableDefinitions.selectedPhotos = nil
	tableDefinitions.space_user = nil
	
	return tableDefinitions
end

--------------------------------------------------------------------------------

function LriCatalogIndexDefinitions.indexDefinitions( indexDefinitions )

	indexDefinitions.profileRegistration_uniqueId = kProfileRegistration_uniqueId
	indexDefinitions.profileRegistration_fileName = kProfileRegistration_fileName
	indexDefinitions.assetProfileLinks_uniqueId = kAssetProfileLinks_uniqueId
	
	return indexDefinitions
	
end

--==============================================================================

local tableDefinitions = {}
local indexDefinitions = {}

--------------------------------------------------------------------------------

local LrmIndexStoreSchema = {
	
	WFRxIndexStoreSchema,
		--superclass
}

function LrmIndexStoreSchema:addTable( args )

	if args.updateLazily ~= true and not args.isManualTable then
		args.updateLazily = false
	end
	
	self:WFRxIndexStoreSchema_addTable( args )
	
end


local LrmIndexStoreSchemaClass = Rx.Class( 'WFRxIndexStoreSchema', LrmIndexStoreSchema )

local indexStoreSchema = LrmIndexStoreSchemaClass( )

indexStoreSchema:addDatabaseInitRoutine( require 'WFSQLiteSetLogicImp'.registerSetModule )

require 'OzBinaryUploadUtils'.filterIndexDefinitions( tableDefinitions, indexDefinitions )

--------------------------------------------------------------------------------

LriCatalogIndexDefinitions.tableDefinitions( tableDefinitions )
LriCatalogIndexDefinitions.indexDefinitions( indexDefinitions )

--------------------------------------------------------------------------------

for k, definition in pairs( tableDefinitions ) do
	indexStoreSchema:addTable( definition )
end

for k, definition in pairs( indexDefinitions ) do
	indexStoreSchema:addIndex( definition )
end

local function getIndexSchemaGenerators(  )
	local indexDefinitions = { 
		'LriCoreInfoIndexDefinitions',
		'LrmAlbumsIndexDefinitions',
		'LriCacheIndexDefinitions',
		'LrFavoritesIndexDefinitions',
		'LrCommentsIndexDefinitions',
		'LrmKeywordsIndexDefinitions',
		'LriPersonsIndexDefinitions',
		'LrmPresetsProfilesIndexDefinitions',
		'LrmTrashIndexDefinitions'
		}
	
	return indexDefinitions
		
end

--------------------------------------------------------------------------------

function LriCatalogIndexDefinitions.generateIndexSchema( indexStoreSchema )

	local indexSchemaGenerators = getIndexSchemaGenerators()
	
	for k, generator in pairs( indexSchemaGenerators ) do
		require ( generator ).generateIndexSchema( indexStoreSchema )
	end
	
end

--------------------------------------------------------------------------------

LriCatalogIndexDefinitions.generateIndexSchema ( indexStoreSchema )

--------------------------------------------------------------------------------

require 'WFCatConflictResolver'.generateIndexSchema( indexStoreSchema )
require 'WFCatErrorHandler'.generateIndexSchema( indexStoreSchema )
require 'WFCatStorage'.generateIndexSchema( indexStoreSchema )
require 'WFCatCollaboration'.generateIndexSchema( indexStoreSchema )
require 'WFCatCollaborator'.generateIndexSchema( indexStoreSchema )

--------------------------------------------------------------------------------

LriCatalogIndexDefinitions.indexStoreSchema = indexStoreSchema

--------------------------------------------------------------------------------

return LriCatalogIndexDefinitions
