--[[----------------------------------------------------------------------------

LriCoreInfoIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'

--------------------------------------------------------------------------------

local LrmCoreInfoIndexHelper = require "LrmCoreInfoIndexHelper"
local LriAssetUtils = require "LriAssetUtils"

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--

-- @mark LriCoreInfoIndexDefinitions namespace

local LriCoreInfoIndexDefinitions = {}

--------------------------------------------------------------------------------

function LriCoreInfoIndexDefinitions.assetIsDisplayable( docValue )

	local docType = docValue.type
	local subtype = docValue.subtype
	
	if docType ~= 'asset' then return false end
	
	if not ( subtype == 'image' or subtype == 'video' ) then return false end
	
	-- Include if we can get a thumbnail from Oz.
					
	return LrmCoreInfoIndexHelper.assetHasMasterOrProxy( docValue )

end

--------------------------------------------------------------------------------

function LriCoreInfoIndexDefinitions.getFavoritedCount( docValue )

	local favoritedCount = 0
	local favorites = docValue and docValue.favorites

	if favorites then
		for k, v in pairs( favorites ) do
			if v then
				favoritedCount = favoritedCount + 1
			end
		end
	end

	return favoritedCount

end

--------------------------------------------------------------------------------

function LriCoreInfoIndexDefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )
	
	local coreInfoCommonFields = LrmCoreInfoIndexHelper.getCoreInfoCommonFields()
	
	local fields = {
		"favoritedCount", "importedBy", "developUserUpdated", "userOrientation", "hasEmbeddedDepthMap", 
		"sha256", "importSourceCameraName", "importSourceLensName", "flashFired", "lens", 
		"isMergedPanorama", 'isCapturedWithAdobeCamera', 'tmp_hasImportSettings', 'syncStatus' 
	}
	
	for _, field in pairs( coreInfoCommonFields ) do
		fields[ #fields+1 ] = field
	end

	indexStoreSchema:addTable {
	
		-- Instead of specifying CREATE TABLE SQL, in order to use
		-- niceToHave, you must specify the field names and let the index
		-- store generate the SQL
		-- jrechtie 2018/09/26 The niceToHave field mentioned above is gone for now.

		tableName = "coreInfo",
		
		fields = fields,
		
		COLLATE = {
			eventName = 'i18n_collate',
		},
		
		docValueFilter = function( docValue )
			return (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset'
		end,

		map = function( docValue, emit )

			if docValue._deleted then
				return
			end
			if not LriCoreInfoIndexDefinitions.assetIsDisplayable( docValue ) then
				return
			end
				
			local favoritedCount = LriCoreInfoIndexDefinitions.getFavoritedCount( docValue )
			
			local importedBy
			local hasEmbeddedDepthMap
			local sha256
			
			local importSource = docValue.importSource
			
			if importSource then
			
				importedBy = importSource.importedBy or ''
				hasEmbeddedDepthMap = importSource.hasEmbeddedDepthMap
				sha256 = importSource.sha256

			end

			local importSourceCameraName, importSourceLensName = LriAssetUtils.getImportSourceCameraAndLensName( docValue )
			
			local developUserUpdated = nil
			local userOrientation = nil

			local develop = derefOpt( docValue, 'develop' )
			if develop then
				if develop.userUpdated then
					developUserUpdated = develop.userUpdated
				end
				if develop.userOrientation then
					userOrientation = develop.userOrientation
				end
			end
			
			
			local xmp = LriAssetUtils.getXmp( docValue )
			
			local isMergedPanorama = derefOpt( docValue, 'xmp', 'aux', 'IsMergedPanorama' )
			
			local isCapturedWithAdobeCamera = LriAssetUtils.getCapturedWithAdobeCamera( docValue )

			-- jrechtie, 2018/07/14 temporary field for migrating the importSettings. Will be removed with the first forced reindex after the migration.
			local tmp_hasImportSettings = derefOpt( docValue, '_localOnly', 'lrm_customData', 'importSettings' )
			tmp_hasImportSettings = ( tmp_hasImportSettings and type( tmp_hasImportSettings ) == 'table' ) and 1 or 0

			-- jrechtie, 2018/07/17 syncStatus: not yet implemented. Here as a placeholder for future changes. LrmAssetutils.getSyncStatus() returns
			-- nil for now.
			local syncStatus = LriAssetUtils.getSyncStatus( docValue )

			local nCoreInfoCommonFields, coreInfoCommonFieldsValue = LrmCoreInfoIndexHelper.getCoreInfoCommonFieldsValue( docValue )
			
			emit( favoritedCount, importedBy, developUserUpdated, userOrientation, hasEmbeddedDepthMap,
					sha256, importSourceCameraName, importSourceLensName, xmp.flashFired, xmp.lens,
					isMergedPanorama, isCapturedWithAdobeCamera, tmp_hasImportSettings, syncStatus,
					unpack( coreInfoCommonFieldsValue, 1, nCoreInfoCommonFields )
			 )
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_importedBy ON coreInfo( importedBy, captureDate, docId )]]

	-- Grid sorting criteria which can benefit from from an index.
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_captureDate ON coreInfo( captureDate )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_importTimestamp ON coreInfo( importTimestamp )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_modifiedDate ON coreInfo( modifiedDate, captureDate, fileName COLLATE i18n_collate, docId )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_fileName ON coreInfo( fileName COLLATE i18n_collate  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_developUserUpdated ON coreInfo( developUserUpdated  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_sha256 ON coreInfo( sha256  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_localUrl ON coreInfo( localUrl  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_isCapturedWithAdobeCamera ON coreInfo( isCapturedWithAdobeCamera )]]
	-- Note: an index on column 'rating' actually increases sorting time significantly.
	-- This is probably due to it being a "low quality index" - one with more than 10-20 rows having the same value in left most column.

	-- Note: an index on columns 'flag, docId' actually increases "filter for flag" in time in albums significantly.
	-- This is due to it being a "low quality index" - with only three possible values for flag, where one value is heavily
	-- dominating for typical catalogs: "unflagged".
	-- In a complex query with several filters, the query optimzer might choose the flag_docId index as first filter
	-- and then only the albumid in the album_asset table, leading to unneccessary long query times.

end

--------------------------------------------------------------------------------

return LriCoreInfoIndexDefinitions
