--[[----------------------------------------------------------------------------

LriOzPurgeAgent.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the content of a single catalog's data cache and periodically
purges it.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- WF.core.rxlua
-- local WFRxScheduler = require 'WFRxScheduler'

local WFArguments = require 'WFArguments'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local KSIdGenerator = require "KSIdGenerator"

	-- WF.os.prefs
-- local KSPrefs = require 'KSPrefs'

	-- WF.oz.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'

	-- WF.oz_client.sync_engine
-- local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'

local WFCatTrashUtils = require 'WFCatTrashUtils'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

local LrmConfig = require "LrmConfig"
local WFCapabilities = require 'WFCapabilities'

--------------------------------------------------------------------------------

LrmConfig.registerBinaryStrategyCheck()

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

local cachePollingInterval = 40	-- seconds; feel free to tune

local minimumFreeSpace = WFOzPurgeSettings.minimumFreeSpace()

local ENABLE_CLOUD_TRASH = WFCapabilities.getCapability( "WFCatDelete.enableCloudTrash" ) and true or false

local logger = WFLogger 'LriOzPurgeAgent'
	-- :setLevel 'DEBUG'

local derefOpt = KSTableUtils.derefOpt

--------------------------------------------------------------------------------

local getDiskSpaceNeeded = function( self )

	local fileSystemFreeSpace, _ -- fileSystemSize
		= WFFileUtils.freeSpaceAtPath( KSPathUtils.getStandardFilePath( 'carouselDocuments' ) )

	local requiredSpace = math.max(0, minimumFreeSpace - fileSystemFreeSpace)

	return requiredSpace
end	


--============================================================================--
--[[!
	\brief A \c LriOzPurgeAgent monitors the content of a single catalog's data
	cache and periodically purges it to make additional disk space available.
	
	It should be used only by LrmSyncClientImp.	
--]]

-- @mark LriOzPurgeAgent

local purgeAgentMethods = {
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (OzIndexStore)
	
	_catalogBasePathStream = false,
		-- (Observable< String >) catalog base path
	
	_dataCacheStream = false,
		-- (Observable< KSDataCache >) data cache for this doc store
	
	_dataCachePathStream = false,
		-- (Observable< string >) path to cache
	
	_purgeUserUpdated = false,
		-- An IndexStore query listing purgeable assets by userUpdated
	
	_purgeAll = false,
		-- An IndexStore query listing all purgeable	
	
	_purgeOnDeleteChain = false,
		-- the reactive chain that is executed on purgeOnDelete (exposed for testing)
	
	_purgeOnDelete = false,
		-- Subscription to the docStore's document deleter
	
	_checkCacheKeysFrom = false,
		-- (false or string) where do we start cross-checking cache keys?

	_warnedMissingEntries = false,
		-- (table) entries that we've warned about (so we don't keep bleating about them)

	_binaryStrategy = false,
		-- (string) LrmConfig
		
	_purgeSub = false,
		-- A subscription to _purgeUserUpdated used to continously purge unused renditions.

	private = {
		
		-- The following are methods are internal, but are exposed for testing.
		_getDiskSpaceNeeded = getDiskSpaceNeeded
	},

	WFCloseSupport.addCloseSupportMethods,

}


--------------------------------------------------------------------------------

function purgeAgentMethods:sendDiagnosticLogData()

	return {

		name = "Lrm Purge Agent",
		id = 'LriOzPurgeAgent',
		weight = 521,

		body = {
			binaryStrategy = self._binaryStrategy,	
		}
	}
	
end

--------------------------------------------------------------------------------

local function purgeableAlbumAssetListSql()

	return [[
		SELECT docId, renditionName, path, size 
		FROM cacheReferences 
		WHERE 
			cacheReferences.hasUploaded == 1
			AND cacheReferences.renditionName != 'thumbnail2x'
			AND docId IN ( 
				SELECT 
					album_asset.assetId 
				FROM 
					album_asset
				INNER JOIN 
					albums 
				ON 
					album_asset.albumId == albums.docId 
				WHERE 
					album_asset.albumId == ?)
			AND (docId NOT IN ( 
					SELECT 
					   album_asset.assetId 
					FROM 
					   album_asset 
					INNER JOIN 
					   albums 
					ON 
						   album_asset.albumId == albums.docId 
					WHERE 
						   albums.useOffline == 1 
					)
				)
		   LIMIT 100 ]]
end
 
 --------------------------------------------------------------------------------

 local function purgeUserUpdatedSql()
 
	 return [[ 
		SELECT docId, renditionName, path, size 
		FROM cacheReferences
			WHERE userUpdated < ( SELECT max( userUpdated ) FROM cacheReferences )
				AND cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != 'thumbnail2x'
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			ORDER BY userUpdated ASC, size DESC LIMIT 100 ]]
end

 --------------------------------------------------------------------------------

local function purgeAllSql()

	return [[ 
		SELECT docId, renditionName, path, size 
		FROM cacheReferences 
			WHERE cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != 'thumbnail2x'
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			LIMIT 100 ]]

end

--------------------------------------------------------------------------------

local function setBinaryStrategy( self, binaryStrategy )

	if self._binaryStrategy ~= binaryStrategy then

		self._binaryStrategy = binaryStrategy

		self._purgeUserUpdated = self._indexStore:query {
			sql = purgeUserUpdatedSql(),
			calm = cachePollingInterval
		}

		self._purgeAll = self._indexStore:query {
			sql = purgeAllSql(),
			queryId = 'purgeAll',
		}

	end

end

--------------------------------------------------------------------------------

local function removeAssets( self, sql, ... )

	local queryId = 'removeAssets' .. KSIdGenerator.localCounter()

	return self._indexStore:query{
			sql = sql,
			quiesceUntilUpToDate = true,
			queryId = queryId,
			...
		}
		:takeWhile( function( docIds )
			return next( docIds )
		end )
		:flatMap( function( docIds )
			
			local bulkOverlay = { label = 'oz' }

			for i, r in ipairs( docIds ) do
		
				table.insert(bulkOverlay, {
					_id = r[ 1 ],
					_deleted = true
				} )
			end
					
			return self._docStore:bulkApplyOverlay( bulkOverlay )
				:map( function( revs )
					return #revs
				end )	
		end )
		:tap(function()
		
			self._indexStore:requestFromQuery( queryId, 100 )

		end )
		:sum()
end 

--------------------------------------------------------------------------------

local function setupPurgeOnDelete( self )

	self._purgeOnDeleteChain = self._docStore:deletedDocStream()

		:filter( function( docValue )
			return docValue.type == "asset" or docValue.type == 'album' or docValue.type == 'space'
		end )
				
		:flatMap( function( docValue )
			
			if docValue.type == 'asset' then

				return self._catalogBasePathStream
					:flatMap( function( catalogBasePath )
						
						local hasTrashOrRestoreAsset = WFCatTrashUtils.trashAssetInfo( docValue ) ~= nil or WFCatTrashUtils.restoredToInfo( docValue ) ~= nil
						
						if ENABLE_CLOUD_TRASH and hasTrashOrRestoreAsset then		
							-- Leave purge Document to blobProvider
							return Rx.Observable.just( docValue )
						else 
							return Rx.Observable.just( self:purgeDeletedDocument( docValue, catalogBasePath ) )
						end
						
					end )
			
			elseif docValue.type == 'album' then

				return self:purgeForAlbum( docValue._id )
					:flatMap( function ()
				
						return removeAssets( self,
							[[ SELECT docId FROM album_asset WHERE albumID = ? LIMIT 100]],
							docValue._id )

					end )
					:defaultIfEmpty( docValue )
					:map( function()
						return docValue
					end )
			
			elseif docValue.type == 'space' then 			

				return removeAssets( self,
					[[SELECT docId FROM comment WHERE spaceId = ?
						UNION SELECT docId FROM spacefavorite where spaceId = ? LIMIT 100]],
						docValue._id,
						docValue._id)
					:map( function()
						return docValue
					end )

			end

		end )
		:pluck( '_id' )
		:shareReplay( 1 )

	self._purgeOnDelete = self._purgeOnDeleteChain:subscribe()
	
	-- value produced is the number of entries removed from the cache
end 

--------------------------------------------------------------------------------

--[[!
	\brief Initialize a purge agent for a specific catalog.
	
	\param catalogService (OzCatalogServiceViaWFCatalog) catalog service
	\param binaryStrategy (LrmConfig) prefer proxy or master
--]]

function purgeAgentMethods:__init( catalogService, binaryStrategy )

	check( 'LriOzPurgeAgent', 'OzCatalogService', '?binaryStrategy' )
	
	self._docStore = catalogService:docStore()
	self._indexStore = catalogService:indexStore()
	
	self._dataCacheStream = catalogService:getDataCache()
		:shareReplay( 1 )

	self._dataCachePathStream = catalogService:getDataCachePath()
	self._dataCachePathStream:call( runUntilDone )
	
	self._catalogBasePathStream = catalogService:getCatalogBasePath()
		:shareReplay( 1 )
	
	self._warnedMissingEntries = {}
	
	if binaryStrategy then
		setBinaryStrategy( self, binaryStrategy )
	end

	setupPurgeOnDelete( self )

	WFDiagnosticLog.addLogProvider( self )
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:close()

	local purgeOnDelete = self._purgeOnDelete
	self._purgeOnDelete = nil
	
	if purgeOnDelete then
		purgeOnDelete:dispose()
	end

	self:stopContinuousPurge()

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeDeletedDocument( docValue, catalogDir )

	local renditionNames = { 'master', 'proxy', 'preview', 'thumbnail2x', 'xmp_develop' }
	
	for _, renditionName in ipairs( renditionNames ) do
	
		local renditionPath = derefOpt( docValue, '_localOnly', 'files', renditionName, 'path' )
		if renditionPath then
			WFFileUtils.delete( KSPathUtils.child( catalogDir, renditionPath ) )
		end
	
	end

	return docValue

end

--------------------------------------------------------------------------------

--[[

\brief prepare purging of documents

\param victims as an array of { docId, renditionName, path, size }
	   (the output of an index store query)
\param spaceNeeded: spaceNeeded to satisfy minimumFreeSpace threshold or
		-1 for no purge limit

\returns a table with
	- docs:  changed documents
	- removeFiles: files relative to the Documents dir that can be removed
	- spaceReclaimed: space reclaimed
	- purged: number of purged documents
]]--

local function preparePurge( victims, spaceNeeded )

	logger:infof( "preparePurge\nvictims = %s\nspaceNeeded = %s",
				   victims, spaceNeeded )

	local result = {
		docs = {},
		removeFiles = {},
		purged = 0,
		spaceReclaimed = 0
	}

	if spaceNeeded == 0 or not next(victims) then
		return result -- we're done
	end

	for i, victim in ipairs( victims ) do

		local docId = victim[ 1 ]

		local renditionName = victim[ 2 ]
			-- Hard-coded indices are based on indexStore queries purgeAllSql, purgeUserUpdatedSql 
			-- and purgeableAlbumAssetListSql
		
		local overlay = result.docs[ docId ] or {
			_id = docId
		}

		KSTableUtils.setNested( overlay, { _delete = true }, '_localOnly', 'files', renditionName )

		local path = victim[ 3 ]

		result.removeFiles[ #result.removeFiles + 1 ] = path
	
		result.docs[ docId ] = overlay
	
		result.spaceReclaimed = result.spaceReclaimed + victim[ 4 ]
		result.purged = result.purged + 1

		if ( spaceNeeded ~= -1 ) and ( result.spaceReclaimed >= spaceNeeded ) then
			return result
		end

	end
			
	return result

end


--------------------------------------------------------------------------------

local function purgeVictims(self, victims, spaceNeeded )

	local purged = 0

	local batch = preparePurge( victims, spaceNeeded )
	
	local overlays = KSTableUtils.extractValues( batch.docs )
	
	if #overlays == 0 then
		return Rx.Observable.just( 0, spaceNeeded )
	end
	
	return self._catalogBasePathStream
		:flatMap( function( catalogBasePath )
		
			return self._docStore:bulkApplyOverlay( overlays )
				:map( function()
		
					purged = purged + batch.purged
					if spaceNeeded ~= -1 then
						spaceNeeded = math.max(0, spaceNeeded - batch.spaceReclaimed)
					end
		
					-- remove the files
		
					for _, fileName in ipairs( batch.removeFiles ) do
						local absPath = KSPathUtils.child( catalogBasePath, fileName )
						WFFileUtils.delete( absPath )
					end
			
					return purged, spaceNeeded
									
				end )

		end )
end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeUntilDone( query, spaceNeeded, queryId )

	if spaceNeeded == 0 then
		return Rx.Observable.just( 0 )
	end

	return query
	
		:takeWhile( function( victims )
		
			return next( victims )
		
		end )
	
		:flatMap( function( victims )
						
			return purgeVictims( self, victims, spaceNeeded )
			
		end )

		:tap(function()
		
			self._indexStore:requestFromQuery( queryId, 100 )

		end )
		
		:map( function (_purged, _spaceNeeded)

			spaceNeeded = _spaceNeeded
	
			return _purged
			
		end )
		
		:sum()
		
end

--------------------------------------------------------------------------------

--[[!
	\brief Triggers an immediate purge cycle.
]]--

function purgeAgentMethods:purgeImmediately( binaryStrategy, noPurgeLimit )

	check( 'LriOzPurgeAgent', '?binaryStrategy', '?boolean' )

	if binaryStrategy then
		setBinaryStrategy( self, binaryStrategy )
	end

	local spaceRequired = self.private._getDiskSpaceNeeded()
	if noPurgeLimit then
		spaceRequired = -1
	end

	return self:purgeUntilDone( self._purgeAll, spaceRequired, 'purgeAll' )
		
end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeForAlbum( albumId, binaryStrategy, noPurgeLimit )
	
	check( 'LriOzPurgeAgent', 'uuid', '?binaryStrategy', '?boolean' )

	logger:debugf( "purgeForAlbum, %s", albumId )
	
	local indexStore = self._indexStore
	local docStore = self._docStore
	
	local albumDoc = docStore:docValue( albumId )
	
	local spaceRequired = minimumFreeSpace
	if noPurgeLimit then
		spaceRequired = -1
	end

	-- IMPORTANT!!
	--
	-- Verify offline use is still disabled for the album otherwise we are gonna ddos Oz
	-- as the PurgeAgent will try to remove files locally whereas the BinaryDownloader 
	-- will download again
	
	local result = albumDoc
		:flatMap( function( docValue )
	
			local useOffline = derefOpt( docValue, '_localOnly', 'useOffline' )
			
			if useOffline == false then
			
				return indexStore:query { 
						sql = purgeableAlbumAssetListSql(),
						quiesceUntilUpToDate = true,
						queryId = 'purgeAlbum',
						albumId
					}
				
			else
			
				return Rx.Observable.just( {} )
				
			end
			
		end )
	
	return self:purgeUntilDone( result, spaceRequired, 'purgeAlbum' )
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:stopContinuousPurge()

	check( 'LriOzPurgeAgent' )

	if self._purgeSub then
		self._purgeSub:dispose()
		self._purgeSub = nil
	end
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:startContinuousPurge( binaryStrategy )

	check( 'LriOzPurgeAgent', 'binaryStrategy' )

	if self._purgeSub and binaryStrategy == self._binaryStrategy then
		-- if we're already running and the binary strategy hasn't changed, we're done
		return
	end

	if self._purgeSub then
		self:stopContinuousPurge()
	end

	setBinaryStrategy( self, binaryStrategy )

	self._purgeSub = self._purgeUserUpdated
		:flatMap( function( potentialVictims )
			
			local spaceNeeded = self.private._getDiskSpaceNeeded()
					
			return purgeVictims( self, potentialVictims, spaceNeeded )
		
		end )
		:subscribe()

	print( self._purgeSub )

end


return Rx.Class 'LriOzPurgeAgent'( purgeAgentMethods )
