--[[----------------------------------------------------------------------------

LriPersonsIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )

--============================================================================--

-- @mark LriPersonsIndexDefinitions namespace

local LriPersonsIndexDefinitions = {}

--------------------------------------------------------------------------------

function LriPersonsIndexDefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE persons( personId, name, hidden, primaryId, cover, thumbnail_url )]],

		useLocalId = true,
		hasOneRowPerDocument = true,
		
		updateLazily = false,
		
		docValueFilter = function( docValue )
			return docValue.subtype == 'person' and docValue.type == 'asset'
		end,

		map = function( docValue, emit )

			if docValue._deleted then
				return
			end

			if docValue.subtype == 'person' and docValue.type == 'asset' then

				local personId = derefOpt( docValue, '_id' )
				local thumbnail_href = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/rendition_type/thumbnail2x', 'href' )
				local name = derefOpt( docValue, 'name' ) or ''
				local hidden = derefOpt( docValue, 'hidden' ) or false
				local primaryId = derefOpt( docValue, 'primary' ) or ''
				local cover = derefOpt( docValue, 'cover', 'id' ) or ''

				emit( personId, name, hidden, primaryId, cover, thumbnail_href )

			end

		end,
	}

	indexStoreSchema:addIndex [[CREATE INDEX persons_personId_name ON persons( personId, name, hidden, primaryId )]]

	--------------------------------------------------------------------------------

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE faces( faceId, assetId, personId, person_confirmed, x_axis, y_axis, width, height )]],

		useLocalId = true,
		hasOneRowPerDocument = false,
		
		updateLazily = false,
		
		docValueFilter = function( docValue )
			return (docValue.subtype == 'video' or docValue.subtype == 'image') and docValue.type == 'asset'
		end,
		
		map = function( docValue, emit )

			if docValue._deleted then
				return
			end

			if (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset' then

				local regionList = derefOpt( docValue, 'xmp','mwg-rs', 'Regions', 'RegionList' )

				for k, region in pairs( regionList or {} ) do

					if region.Type == "Face" then

						local area = derefOpt( region, 'Area' )

						local people = derefOpt( region, 'Extensions', 'adobe-fi', 'People' )

						for personId, person in pairs( people or {} ) do
							if not person.Rejected then
								emit( k, docValue._id, personId, person.Confirmed and true or false, area.x, area.y, area.w, area.h )
							end
						end
					end
				end

			end

		end,
	}

	indexStoreSchema:addIndex [[CREATE INDEX faces_confirmed_asset_person ON faces( person_confirmed, assetId, personId )]]

end

--------------------------------------------------------------------------------

return LriPersonsIndexDefinitions
