--[[----------------------------------------------------------------------------

LrmCoreInfoIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'

--------------------------------------------------------------------------------

local LrmCoreInfoIndexHelper = require "LrmCoreInfoIndexHelper"
local LrmAssetUtils = require "LrmAssetUtils"

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--

-- @mark LrmCoreInfoIndexDefinitions namespace

local LrmCoreInfoIndexDefinitions = {}

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.assetIsDisplayable( docValue )

	local docType = docValue.type
	local subtype = docValue.subtype

	if WFPlatformInfo.isAndroid then
		local video = derefOpt( docValue, 'video' )
		if video or subtype == 'video' then return false end
	end

	if docType ~= 'asset' then return false end
	
	if not ( subtype == 'image' or subtype == 'video' ) then return false end

	-- Include if we can get a thumbnail from Oz.

	return LrmCoreInfoIndexHelper.assetHasMasterOrProxy( docValue )

end

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.hasThumbnail( docValue )

	if derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/rendition_type/thumbnail2x', 'href' ) then
		return true
	end
                               	  
	if derefOpt( docValue, '_localOnly', 'thumbnail2x', 'md5' ) then
		return true
	end

	return false
	
end   

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.coreInfoFromDocValue( docValue )

	local importSource = derefOpt( docValue, 'importSource' )
	local width = importSource and importSource.originalWidth
	local height = importSource and importSource.originalHeight
	
	-- Cropped width and height win.
	
	local develop = docValue.develop
	
	if develop and develop.croppedWidth and develop.croppedHeight then
		width = develop.croppedWidth
		height = develop.croppedHeight
	end
	
	-- Local only thumbnail wins.
	-- No, because this causes a lot of extra computation as large
	-- grid view results are re-sent when one thumbnail is loaded
	-- arahn 05/25/2013
	local thumbnail = derefOpt( docValue, '_localOnly', 'thumbnail2x' )
	
	if width == nil and height == nil and thumbnail and thumbnail.referenceWidth and thumbnail.referenceHeight then
		width = thumbnail.referenceWidth
		height = thumbnail.referenceHeight
	end

	local aspectRatio

	if not docValue._deleted
		and type( width ) == 'number' and width > 0
		and type( height ) == 'number' and height > 0
	then
		aspectRatio = width / height
	end
	
	local thumbnailMD5 = thumbnail and thumbnail.md5
	
	-- FIXME [ngautam 2011-12-13]: Need to replace string.lower by
	-- localization-friendly string conversion.

	local favoritedCount = 0
	local favorites = docValue and docValue.favorites

	if favorites then
		for k, v in pairs( favorites ) do
			if v then
				favoritedCount = favoritedCount + 1
			end
		end
	end

	return aspectRatio, thumbnailMD5, favoritedCount

end

function LrmCoreInfoIndexDefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	--[[ TODO [immisch 2018-05-25]: for the next big reindexing (after the BlobProvider is activated)
		these fields should be removed:
		- thumbnailMD5
		- hasThumbnail
	]]--

	indexStoreSchema:addTable {
	
		-- Instead of specifying CREATE TABLE SQL, in order to use
		-- niceToHave, you must specify the field names and let the index
		-- store generate the SQL
		tableName = "coreInfo",
		
		fields = {
			"aspectRatio", "captureDate", "displayDate", "importTimestamp",
			"copyTimestamp", "thumbnailMD5", "favoritedCount", "importedBy", "contentType", "duration", "developUserUpdated",
			"wasDeveloped", "croppedWidth", "croppedHeight", "fileName", "modifiedDate", "userDefined",
			"width", "height", "sha256", "apertureValue", "isoSpeedRatings", "exposureTime", "cameraMaker",
			"cameraModel", "flag", "rating", "copyright", "isVideo", "localUrl", "fnumber",
			"isMergedHDR", 'isCapturedWithAdobeCamera', 'hasThumbnail', 'importedByUser' },
		
		COLLATE = {
			eventName = 'i18n_collate',
		},
		
		docValueFilter = function( docValue )
			return (docValue.subtype == 'image' or docValue.subtype == 'video') and docValue.type == 'asset'
		end,
		
		-- query won't send updates for changes that only affect niceToHave
		-- columns
		niceToHave = {
			"thumbnailMD5",
		},

		map = function( docValue, emit )

			if docValue._deleted then
				return
			end
			if not LrmCoreInfoIndexDefinitions.assetIsDisplayable( docValue ) then
				return
			end
			
			local aspectRatio, thumbnailMD5, favoritedCount = LrmCoreInfoIndexDefinitions.coreInfoFromDocValue( docValue )

			local hasThumb = LrmCoreInfoIndexDefinitions.hasThumbnail( docValue )

			local displayDate
			local userDefined
			local importTimestamp
			local fileName
			local width
			local height
			local copyTimestamp
			local importedBy
			local hasEmbeddedDepthMap
			local sha256
			
			local importSource = docValue.importSource
			
			if importSource then
				
				fileName = importSource.fileName
				importTimestamp = importSource.importTimestamp
				copyTimestamp = importSource.copyTimestamp
				importedBy = importSource.importedBy or ''
				width = importSource.originalWidth
				height = importSource.originalHeight
				hasEmbeddedDepthMap = importSource.hasEmbeddedDepthMap
				sha256 = importSource.sha256

			end

			local importSourceCameraName, importSourceLensName = LrmAssetUtils.getImportSourceCameraAndLensName( docValue )
			
			-- Capture date may have a "Z" appended to the end, signifying that the
			-- timestamp is in UTC (20 chars). Otherwise, the timestamp is treated as
			-- local time (19 chars). Either way, we just care about getting the
			-- YYYY-MM-DD from the beginning of the string.

			if docValue.captureDate and string.len( docValue.captureDate ) > 10 then
				displayDate = string.replace( string.sub( docValue.captureDate, 1, 10 ), ":", "-" )
			end

			local captureDate = docValue.captureDate
			local modifiedDate = docValue.userUpdated or docValue.changedAtTime or captureDate

			local contentTypeNumber = LrmAssetUtils.getContentTypeNumber( docValue )
			
			-- Cropped width and height win.
			local croppedWidth = width
			local croppedHeight = height

			local developUserUpdated = nil
			local userOrientation = nil

			local develop = derefOpt( docValue, 'develop' )
			if develop then
				if develop.croppedWidth and develop.croppedHeight then
					croppedWidth = develop.croppedWidth
					croppedHeight = develop.croppedHeight
				end
				if develop.userUpdated then
					developUserUpdated = develop.userUpdated
				end
				if develop.userOrientation then
					userOrientation = develop.userOrientation
				end
			end
			
			local isVideo = docValue.subtype == 'video'
		
			local durationRatio = derefOpt( docValue, 'video', 'duration' )
			local duration = nil
			if type( durationRatio ) == 'table' then
				duration = math.ceil( ( durationRatio[ 1 ] or 0 ) / ( durationRatio[ 2 ] or 1 ) )
			end
			
			local rating = LrmAssetUtils.getDefaultAssetRating( docValue )
			local flag = LrmAssetUtils.getDefaultAssetReview( docValue )
			local xmp = LrmAssetUtils.getXmp( docValue )
			
			local wasDeveloped = LrmAssetUtils.isDeveloped( docValue )
			local localUrl = derefOpt( docValue, '_localOnly', 'localUrl' )

			local isMergedHDR = derefOpt( docValue, 'xmp', 'aux', 'IsMergedHDR' )
			local isMergedPanorama = derefOpt( docValue, 'xmp', 'aux', 'IsMergedPanorama' )

			local isCapturedWithAdobeCamera = derefOpt( docValue, 'xmp', 'xmp', 'CreatorTool' )
			isCapturedWithAdobeCamera = isCapturedWithAdobeCamera and type( isCapturedWithAdobeCamera ) == 'string' and
				string.find( isCapturedWithAdobeCamera, 'Adobe Camera' ) == 1 or false
				
			local importedByUser = derefOpt( docValue, '_serverProvidedMetadata', 'importedByUser', 'id' ) -- maybe nil!

			emit( 	aspectRatio, captureDate, displayDate, importTimestamp,
					copyTimestamp, thumbnailMD5, favoritedCount, importedBy, contentTypeNumber,
					duration, developUserUpdated, wasDeveloped, croppedWidth, croppedHeight, fileName, modifiedDate, userDefined, width, height,
					sha256, xmp.apertureValue, xmp.isoSpeedRatings, xmp.exposureTime, xmp.cameraMake, xmp.cameraModel, flag, rating,
					xmp.copyright, isVideo, localUrl, xmp.fnumber, isMergedHDR, isCapturedWithAdobeCamera, hasThumb, importedByUser
			 )
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_importedBy ON coreInfo( importedBy, captureDate, docId )]]

	-- Grid sorting criteria which can benefit from from an index.
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_captureDate ON coreInfo( captureDate )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_importTimestamp ON coreInfo( importTimestamp )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_modifiedDate ON coreInfo( modifiedDate, captureDate, fileName COLLATE i18n_collate, docId )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_fileName ON coreInfo( fileName COLLATE i18n_collate  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_developUserUpdated ON coreInfo( developUserUpdated  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_sha256 ON coreInfo( sha256  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_localUrl ON coreInfo( localUrl  )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_isCapturedWithAdobeCamera ON coreInfo( isCapturedWithAdobeCamera )]]
	-- Note: an index on column 'rating' actually increases sorting time significantly.
	-- This is probably due to it being a "low quality index" - one with more than 10-20 rows having the same value in left most column.

	-- Note: an index on columns 'flag, docId' actually increases "filter for flag" in time in albums significantly.
	-- This is due to it being a "low quality index" - with only three possible values for flag, where one value is heavily
	-- dominating for typical catalogs: "unflagged".
	-- In a complex query with several filters, the query optimzer might choose the flag_docId index as first filter
	-- and then only the albumid in the album_asset table, leading to unneccessary long query times.

end

--------------------------------------------------------------------------------

return LrmCoreInfoIndexDefinitions
