--[[----------------------------------------------------------------------------

NmbConnectors.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.os
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- WF.core.utils
local KSDate = require 'KSDate'
local KSIdGenerator = require 'KSIdGenerator'

local KSTableUtils = require 'KSTableUtils'
local WFCatAlbums = require 'WFCatAlbums'
local WFCatUtils = require "WFCatUtils"
local WFCloseSupport = require "WFCloseSupport"
local WFRxScheduler = require "WFRxScheduler"

	-- WF.web
local KSJsonEncoder = require 'KSJsonEncoder'
local KSJsonDecoder = require "KSJsonDecoder"

	-- WF.core.logging
local WFLogger = require 'WFLogger'

--------------------------------------------------------------------------------

local derefOpt = KSTableUtils.derefOpt
local logger = WFLogger 'NmbConnectors'

local testConfig = require "WFConfiguration".connectionConfig

--============================================================================--

local CONNECTOR_CONFIG_LOCAL_DOC = 'connectorConfigCache'

--------------------------------------------------------------------------------

local configUrl = "https://lightroom.adobe.com/api/connectors"
local stageConfigUrl = "https://stage.adobelr.com/api/connectors"

--------------------------------------------------------------------------------

local function distinctTableCompare( _, tbl1, _, tbl2 )

	return KSTableUtils.deepEquals( tbl1, tbl2 )

end


--------------------------------------------------------------------------------
-- returns a list of connectors - can be filtered by status and active state
local function getAllAvailableConnectors( nmbCatalog, optDebug )
	
	return Observable.combineLatest(
		nmbCatalog:docStore():ongoingLocalDocValue( CONNECTOR_CONFIG_LOCAL_DOC ), 
		nmbCatalog:getUniversalCatalogDocumentSetting( 'connections' ) )
		:map( function( connectorDoc, activeConnections )
			
			local connectors = connectorDoc.config or {}
			
			local result = {}
			
			for t,v in pairs( connectors ) do
			
				if v.serviceId then
					if v.status == 'prod' or optDebug then
						local v1 = KSTableUtils.deepCopy( v ) -- need deepcopy so that distinctTableCompare works
						if activeConnections and activeConnections[ v.serviceId ] then
							v1.active = true
						end
						result[ #result + 1 ] = v1
					end
				end
			end
			
			return result
		end )
		:distinctUntilChanged( nil, distinctTableCompare )
		
end

--------------------------------------------------------------------------------

local function getConnectorInfoForServiceId( catalog, maybeServiceId )
	
	return catalog:docStore():ongoingLocalDocValue( CONNECTOR_CONFIG_LOCAL_DOC )
		:map( function( connectorDoc )
		
			local connector
			
			if maybeServiceId and not connectorDoc._empty then
				local connectors = connectorDoc.config or {}
				
				for t, v in pairs( connectors ) do
					if v.serviceId == maybeServiceId then
						connector = v
					end
				end
			end
			
			return connector
		end )
		
end
--------------------------------------------------------------------------------

local function publishAlbum( catalog, projectAlbumId, projectName )
	
	local docStore = catalog:docStore()
		
	return docStore:docValue( projectAlbumId )
		:flatMapLatest( function( projectAlbumDoc )
			if not projectAlbumDoc or projectAlbumDoc._empty then 
				return Observable.throw( string.format( 'Invalid project albums %s', projectAlbumId ) )
			end
			
			local maybeServiceId = derefOpt( projectAlbumDoc, 'publishInfo', 'serviceId' ) 
			return getConnectorInfoForServiceId( catalog, maybeServiceId )
				:map( function( connectorInfo )
					if not connectorInfo then
				 		error( string.format( "Failed to get connector info for %s", maybeServiceId ) )
					end
					
					return projectAlbumDoc, connectorInfo
				end )				
		end )
		:flatMap( function( projectAlbumDoc, connectorInfo )
			
			local docAlbumId = projectAlbumDoc._id
			local projectLocalId = projectAlbumDoc._localId
			local albumName = projectAlbumDoc.name
			local created = projectAlbumDoc.created or projectAlbumDoc.userCreated or
								derefOpt( projectAlbumDoc, '_serverProvidedMetadata', 'created' )
			local publishCreated = derefOpt( projectAlbumDoc, 'publishInfo', 'created' )
			local publishUpdated = derefOpt( projectAlbumDoc, 'publishInfo', 'updated' )
			local coverId = projectAlbumDoc.cover and projectAlbumDoc.cover.id
			local requestUrl = connectorInfo.connectHref
			if not requestUrl then
				requestUrl = connectorInfo.redirectHref
				if requestUrl then
				    -- this is the redirect path - where we will launch a browser to publish
					requestUrl = requestUrl.."?project_id="..projectAlbumId
                    local baseOverlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( projectAlbumId )

                    baseOverlay.publishInfo = {
						-- don't set/update dates, since we haven't published yet
						remoteLinks = {
							edit = connectorInfo.editHref,
							redirect = requestUrl,
						}
					}
					local notificationUrl = derefOpt( projectAlbumDoc, '_localOnly', 'publishNotifications', 'editUrl' )
					if notificationUrl == 'pending' then
						-- remove the local only section for pending publish and, instead, mark for publish
						baseOverlay._localOnly = {
							publishNotifications = {
								serviceId = connectorInfo.serviceId,
								editUrl = requestUrl
							}
						}
					end
                    return docStore:applyOverlay( baseOverlay )
                        :map( function()
							 return { albumId = projectLocalId, editUrl = requestUrl, projectAlbumId = projectAlbumId }
                        end )
                else
                    return Observable.throw( string.format( 'Invalid connector config: no connect or redirect Href for %s', connectorInfo.name ) ) 
                end					
			end
			
			local data = KSJsonEncoder.encodeOzUploadJson( {
				resources = {
					{
						id = docAlbumId,
						type = 'album',
						subtype = projectAlbumDoc.subtype,
						created = created,
						payload = {
							name = albumName,
							cover = coverId and { id = coverId },
						},
						publishInfo = {
							version = 2,
							serviceId = connectorInfo.serviceId,
							-- dates will be null if this is the first publish
							created = publishCreated,
							updated = publishUpdated,
						}
					}
				}
			} )

			local catalogSpec = catalog:ozCatalogSpec()
			return catalogSpec:put( requestUrl, data,
					{ 	requestMethod = 'PUT',
						requestHeaders = {
							{ header = 'Content-Type', value = 'application/json' },
						},
					} )
			:flatMap( function( data, headers )
                if headers and headers.status_code == "201" then
                    local response = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
                    local editUrl = derefOpt( response, docAlbumId, 'editor_url' )
                    -- write Urls and other response data to local project album here
                    local baseOverlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( projectAlbumId )
					baseOverlay.publishInfo = {
 						-- update the publishInfo dates and url to indicate we have published
						created = publishCreated or baseOverlay.userUpdated,
						updated = baseOverlay.userUpdated,
						remoteLinks = {
							edit = editUrl,
						}
					}
					
					local notificationUrl = derefOpt( projectAlbumDoc, '_localOnly', 'publishNotifications', 'editUrl' )
					if notificationUrl == 'pending' then
						-- remove the local only section for pending publish and, instead, mark for publish
						baseOverlay._localOnly = {
							publishNotifications = {
								serviceId = connectorInfo.serviceId,
								editUrl = editUrl
							}
						}
					end
					
                    return docStore:applyOverlay( baseOverlay )
                        :map( function()
							 return { albumId = projectLocalId, editUrl = editUrl, projectAlbumId = projectAlbumId }
                        end )
                else
                    return Observable.throw( string.format( 'Error calling connector service: %s', data ) ) 
                end
			end )
			
       end )
	   :catch( function ()
	   		-- if we had any trouble in this process, make sure we don't automatically keep publishing it.
	   		--WFCatAlbums.removeProjectAlbumPublishNotification( catalog, projectAlbumId ) 
			catalog:docStore():applyOverlay {
				_id = projectAlbumId,
				_localOnly = {
					publishNotifications = {
						_delete = true,
					}
				}
			}
			return Observable.just( { editUrl = 'invalid', projectAlbumId = projectAlbumId } )
	   end )
	   :tapOnError( function( errorMessage )
			logger:errorf( "Create Project Album failed: %s", errorMessage )
		end )
end

--[[ Saved code for direct creation of project album on Oz
local function addAssetsOnOz( catalog, projectId, assetList )

	local catalogSpec = catalog:ozCatalogSpec()
	local addAssetsUrl = string.format( '%s/albums/%s/assets', catalogSpec:catalogHref(), projectId )
	local body = {
		resources =  {
			{
--				id = -- create list of ids here
			},
		},
	}
	return catalogSpec:put( addAssetsUrl, body,
		{ 	requestMethod = 'PUT',
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
			},
		} )
	:tapOnError( function( errorResponse, errorHeaders )
		logger:infof( "Create Project Album failed with %s", errorResponse )
	end )
	:flatMap( function( data, headers )
		if headers and headers.status_code == "201" then
		end
		return Observable.just( nil )
	end )
end

--------------------------------------------------------------------------------

local function createProjectAlbumOnOz( catalog, sourceAlbumId, projectName )

	local projectId = KSIdGenerator.uuidInLowercaseHex()
	local catalogSpec = catalog:ozCatalogSpec()
	
	local requestUrl = string.format( '%s/albums/%s', catalogSpec:catalogHref(), projectId )
	local data = KSJsonEncoder.encodeOzUploadJson(
		{
			type = 'album',
			subtype = 'project',
			name = projectName,
			sourceAlbum = sourceAlbumId and {
				id = sourceAlbumId,
			},
		}
	)
	return catalogSpec:put( requestUrl, data,
			{ 	requestMethod = 'PUT',
				requestHeaders = {
					{ header = 'Content-Type', value = 'application/json' },
				},
			} )
	:tapOnError( function( errorResponse, errorHeaders )
		logger:infof( "Create Project Album failed with %s", errorResponse )
	end )
	:flatMap( function( data, headers )
		if headers and headers.status_code == "201" then
			addAssetsOnOz( catalog, projectId, {} )
			return Observable.just( data )
		else
			return Observable.just( nil )
		end 
	end )
end
 --]]

--------------------------------------------------------------------------------

local function createProjectAlbum( catalog, serviceId, projectName, maybeSourceAlbumFullDocId, maybeCoverId )
	
	local docStore = catalog:docStore()

	local projectId = KSIdGenerator.uuidInLowercaseHex()	

	local projectAlbumDoc = {
		_id = projectId,
		type = 'album',
		subtype = 'project',
		name = projectName,
		sourceAlbum = maybeSourceAlbumFullDocId and {
			id = maybeSourceAlbumFullDocId,
		},
		cover = maybeCoverId and {
			id = maybeCoverId,
		},
		publishInfo = {
			version = 2,
			serviceId = serviceId
		},
		_localOnly = {
			publishNotifications = {
				serviceId = serviceId,
				editUrl = 'pending'
			}
		}
	}
	
	projectAlbumDoc = WFCatUtils.addUserCreatedFields( projectAlbumDoc )
	
	
	
	-- Wait until all of the above operations have completed,
	-- then extract the local doc ID from doc store.
	
	return docStore:setDocValue( projectAlbumDoc )
		:map( function()
			return projectId
		end )
		
end

--------------------------------------------------------------------------------

local function synchronizeAssetsInSourceAlbumToDestinationAlbum( catalog, sourceAlbumId, projectAlbumId )
	
	local assetIdsForSourceAlbumObservable = WFCatAlbums.getAlbumAssetIds( catalog, sourceAlbumId )

	local assetIdsForProjectAlbumObservable = WFCatAlbums.getProjectAlbumAssetIds( catalog, projectAlbumId )

	return Observable.combineLatest( assetIdsForSourceAlbumObservable, assetIdsForProjectAlbumObservable )
		:map( function( sourceAssets, destinationAssets )
			
			local assetIdsToAdd = {}
			
			-- find source assets that are not in destination
			for k, v in pairs( sourceAssets ) do
				if not destinationAssets[ k ] then
					assetIdsToAdd[ #assetIdsToAdd + 1 ] = k
				else
					-- remove it from destination assets, because it's in source
					-- when done, we'll have what's left in destinationAssets
					-- to be removed from the project album
					destinationAssets[ k ] = nil
				end
			end
			
			local assetIdsToRemove = {}
			for k, v in pairs( destinationAssets ) do
				assetIdsToRemove[ #assetIdsToRemove + 1 ] = k
			end
			
			return assetIdsToAdd, assetIdsToRemove
		
		end )
		:flatMap( function( assetIdsToAdd, assetIdsToRemove )
			
			local addAssetsObservable = #assetIdsToAdd > 0 and
										WFCatAlbums.addAssetsToAlbum( catalog, projectAlbumId, assetIdsToAdd ) or
										Observable.just()
										
			local removeAssetsObservable = #assetIdsToRemove > 0 and
							WFCatAlbums.removeAssetsFromAlbum( catalog, projectAlbumId, assetIdsToRemove ) or
							Observable.just()


			return Observable.combineLatest( addAssetsObservable, removeAssetsObservable )
				:takeLast( 0 )
				:concat( Observable.just( assetIdsToAdd, assetIdsToRemove ) )
				
		end )
	
end

--------------------------------------------------------------------------------

local function createConnectionWithAssetIds( catalog, serviceId, assetIds, projectName, maybeSourceAlbumId, maybeCoverId )
	
	return createProjectAlbum( catalog, serviceId, projectName, maybeSourceAlbumId, maybeCoverId )
		:flatMap( function( projectAlbumId )
			return WFCatAlbums.addAssetsToAlbum( catalog, projectAlbumId, assetIds )
				:concat( catalog:docStore():docValue( projectAlbumId ) )
				:takeLast( 1 )
				:map( function( projectAlbum )
					return { albumId = projectAlbum._localId, projectAlbumId = projectAlbum._id }
				end )
		end )
		
end

--------------------------------------------------------------------------------

local function assetIdsForAlbum( catalog, sourceAlbumId )
	
	return  WFCatAlbums.getAlbumAssetIds( catalog, sourceAlbumId )
		:map( function( assetIdsToLocalIds )
			local assetIds = {}
			for docId, localId in pairs( assetIdsToLocalIds ) do
				assetIds[ #assetIds + 1 ] = docId
			end
			return assetIds
		end )

end

--------------------------------------------------------------------------------

local NmbConnectors = {}

--------------------------------------------------------------------------------

function NmbConnectors.createConnectionFromAlbum( catalog, serviceId, sourceAlbumId )
	
	return catalog:docStore():docValue( sourceAlbumId )
		:flatMapLatest( function( sourceAlbumDoc )
			
			local sourceAlbumFullDocId = KSTableUtils.derefOpt( sourceAlbumDoc, '_id' )
			local projectName = KSTableUtils.derefOpt( sourceAlbumDoc, 'name' )
			local coverId = KSTableUtils.derefOpt( sourceAlbumDoc, 'cover', 'id' )
			
			if not sourceAlbumFullDocId or not projectName then
				return Observable.throw( string.format( "source album %s is not a valid album to create a project from", sourceAlbumId ) )
			end
			
			return assetIdsForAlbum( catalog, sourceAlbumId )
				:flatMap( function( assetIds )
					return createConnectionWithAssetIds( catalog, serviceId, assetIds, projectName, sourceAlbumFullDocId, coverId )
				end )
		end )
end

--------------------------------------------------------------------------------

function NmbConnectors.createConnectionFromPhotos( catalog, serviceId, assetIds, maybeProjectName )
	
	-- SPAM [matthewj 2018-08-10]: This really should default on the UI side, but for now
	-- with no UI, we'll populate this here
	maybeProjectName = maybeProjectName or KSDate.politeDateFromTime( KSDate.timeNow(), 'medium', 'short' )
	
	return createConnectionWithAssetIds( catalog, serviceId, assetIds, maybeProjectName )
	
end

--------------------------------------------------------------------------------

function NmbConnectors.updateConnection( catalog, projectAlbumId )
	
	local sourceAlbumId, projectName
	return catalog:docStore():docValue( projectAlbumId )
		:flatMapLatest( function( projectDoc )
			sourceAlbumId = derefOpt( projectDoc, 'sourceAlbum', 'id' )
			projectName = derefOpt( projectDoc, 'name' )
			local serviceId = derefOpt( projectDoc, 'publishInfo', 'serviceId' )
			
			if not serviceId then
				return Observable.throw( "Invalid project or unknown service" )
			end
			
			if sourceAlbumId then
				-- sourceAlbumId may exist, but not for a valid (or yet synced) source album
				-- so try to get the actual source album doc value. If we get one, treat 
				-- synchronize the assets between the two. If we don't, treat the republish
				-- like it's an adhoc for now and do not change it's asset membership.
				return catalog:docStore():docValue( sourceAlbumId )
			else
				return Observable.just()
			end
		end )
		:flatMap( function( maybeSourceAlbumDocValue )
			
			local maybeSynchronizeAssets
			
			if maybeSourceAlbumDocValue and not maybeSourceAlbumDocValue._empty then
				maybeSynchronizeAssets = synchronizeAssetsInSourceAlbumToDestinationAlbum( catalog, sourceAlbumId, projectAlbumId )
			else
				maybeSynchronizeAssets = Observable.just()
			end
				
			return maybeSynchronizeAssets
				:map( function()
					return projectAlbumId, projectName
				end )
				-- SPAM [matthewj 2018-08-10]: Need to do better than a simple 'delay' here as we might end up calling publish
				-- before everything is on Oz
				:delay( 2, WFRxScheduler.sharedScheduler() )
				:flatMap( function( projectAlbumId, projectName )
		
					return publishAlbum( catalog, projectAlbumId, projectName )
		
				end )
		end )
end

--------------------------------------------------------------------------------

function NmbConnectors.getConnectionAttributes( catalog, projectAlbumId )

	return catalog:docStore():docValue( projectAlbumId )
		:flatMapLatest( function( projectDoc )
			local attrs = {}
			if projectDoc and not projectDoc._empty then
				attrs.serviceId = derefOpt( projectDoc, 'publishInfo', 'serviceId' )
				attrs.created = derefOpt( projectDoc, 'publishInfo', 'created' )
				attrs.updated = derefOpt( projectDoc, 'publishInfo', 'updated' )
				attrs.sourceAlbumId = derefOpt( projectDoc, 'sourceAlbum', 'id' )
				attrs.editUrl = derefOpt( projectDoc, 'publishInfo', 'remoteLinks', 'edit' ) or projectDoc.url
				attrs.viewUrl = derefOpt( projectDoc, 'publishInfo', 'remoteLinks', 'view' )
				attrs.viewUrl = derefOpt( projectDoc, 'publishInfo', 'remoteLinks', 'redirect' )
			end
			return Observable.just( attrs )
		end )
end

--------------------------------------------------------------------------------

function NmbConnectors.updateAvailableConnectorsCache( catalog )
	local catalogSpec = catalog:ozCatalogSpec()
	
	local languageCode = WFLocalizationUtils.getLanguageCode()
	local url =  catalogSpec:isStageEnv() and stageConfigUrl or configUrl
	--local WFZString = require "WFZString" -- may need to add this trick to get a working language code in the process...
	--local languageCode = WFZString.fromZString( "$$$/locale_metric/ISO_639_1/LanguageCode=en" )
	if languageCode then
		url = url.."?locale="..languageCode
	end
	return catalogSpec:get( url,
					{ 	requestHeaders = {
							{ header = 'locale', value = languageCode },
						},
					} )
		:map( KSJsonDecoder.decodeWebSafeJson_stripNull )
		:flatMap( function( response, headers )
			local config = KSJsonDecoder.decodeWebSafeJson( response )
			if not config then
				error( 'Unexpected response to connector config request' )
			end
			-- config.sys can include a url to an extra configuration for testing
			if testConfig and testConfig.url then
				local catalogSpec = catalog:ozCatalogSpec()
				return catalogSpec:get( testConfig.url.."?locale="..languageCode, {} )
				:map( function( data, headers )
                	if headers and headers.status_code == "200" then
						local response = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
                    	local serviceId = response.serviceId
						if serviceId then 
							config[ serviceId ] = response
						end
					else
						logger:warnf( "unable to load test connector config ", testConfig.url )
					end
					return config
				end)
			else
				return Observable.just(config)
			end
		end )
		:flatMap( function( config )
			return catalog:docStore():setLocalDocValue( {
				_id = CONNECTOR_CONFIG_LOCAL_DOC, 
				config = config
			} )
		end )
		:tapOnError( function( errorResponse, errorHeaders )
			logger:warnf( "Failed to retrieve updated updated connector config", errorResponse )
		end )
end

--------------------------------------------------------------------------------

function NmbConnectors.addConnectorModels( nmbModelBuilder )
	
	nmbModelBuilder.addCatalogBasedMostRecentValueModel( 'getAllAvailableConnectors', 
			function( wfCatalog, status )
				return getAllAvailableConnectors( wfCatalog, status )
			end )
					
	nmbModelBuilder.addCatalogBasedMostRecentValueModel( 'addActiveConnector',
			function( nmbCatalog, serviceId )
				local newConnectionInfo = {
					[ serviceId ] = {
						created = KSDate.timeInIso8601Format()
					}
				}
				-- SPAM [matthewj 2018-08-01]: This seems a bit clunky the way this API is currently written
				-- ideally, it'd be simpler to add this info without having to think about whether it's an 
				-- overlay or not.  Maybe just pushing this knowledge into NmbConnectors is fine though.
				return nmbCatalog:overlayUniversalCatalogDocumentSetting( 'connections', newConnectionInfo )
			end )
			
	nmbModelBuilder.addCatalogBasedMostRecentValueModel( 'removeActiveConnector', 
			function( nmbCatalog, serviceId )
				local newConnectionInfo = {
					[ serviceId ] = {
						_delete = true
					}
				}
				return nmbCatalog:overlayUniversalCatalogDocumentSetting( 'connections', newConnectionInfo )
			end )
			
	nmbModelBuilder.addCatalogBasedMostRecentValueModel( 'getActiveConnectors',
			function( nmbCatalog )
				return nmbCatalog:getUniversalCatalogDocumentSetting( 'connections' )
			end )

end

--------------------------------------------------------------------------------

local function watchForSyncCompletionToPublishConnectorExternally( wfCatalog )
	
	local projectAlbumObservable =  wfCatalog:indexStore():query( {
			sql = [[SELECT docId, albumId, name, serviceId 
					FROM project_albums
					WHERE pendingPublish = 1]],
			quiesceUntilUpToDate = true,
		} )
		:map( function( rows )
			local projectAlbums = {}
			for _, row in ipairs( rows ) do
				local id = row[ 1 ]
				local projectAlbum = {
					localId = id,
					docId = row[ 2 ],
					name = row[ 3 ],
					serviceId = row[ 4 ],
				}
				projectAlbums[ #projectAlbums + 1] = projectAlbum
			end
			return projectAlbums
		end )
	
	return Observable.combineLatest( projectAlbumObservable, require 'WFCatSyncStatus'.getAllAlbumsFinishedUploading( wfCatalog ) )
		:map( function( projectAlbums, syncStatus ) 
			
			local albumsToPublish = {}
			for _, album in ipairs( projectAlbums ) do
				if syncStatus[ album.docId ] then
					albumsToPublish[ #albumsToPublish + 1 ] = album
				end
			end
			
			return albumsToPublish
		end )
		:distinctUntilChanged()
		:flatMap( Observable.from )
		:flatMap( function( albumToPublish )
			
			return publishAlbum( wfCatalog, albumToPublish.docId, albumToPublish.name )
				
		end )
		
end

--------------------------------------------------------------------------------
-- @mark NmbConnectorPublisher
--------------------------------------------------------------------------------

local NmbConnectorPublisher = {
	
	_connectorPublishSubscription = false,
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------
	
function NmbConnectorPublisher:__init( wfCatalog )

	self._connectorPublishSubscription = watchForSyncCompletionToPublishConnectorExternally( wfCatalog )
		:subscribe( nil,
		function() 
			logger:debugf( "Connector publish subscription received onDone" )
		end,
		function( errorMessage )
			logger:warnf( "Connector publish subscription received onError: %s", errorMessage )
		end )
	

end

--------------------------------------------------------------------------------

function NmbConnectorPublisher:close()
	
	local publishSubscription = self._connectorPublishSubscription
	
	self._connectorPublishSubscription = nil
	
	if publishSubscription then
		publishSubscription:dispose()
	end
	
end

--------------------------------------------------------------------------------

local NmbConnectorPublisher_Class = Rx.Class( 'NmbConnectorPublisher', NmbConnectorPublisher )
	
--------------------------------------------------------------------------------

function NmbConnectors.createConnectorPublisher( wfCatalog )
	
	return NmbConnectorPublisher_Class( wfCatalog )
	
end

--------------------------------------------------------------------------------

return NmbConnectors
