--[[----------------------------------------------------------------------------

PixelProviderChooseRenditionStrategyTask.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wf.core
local WFConfiguration = require 'WFConfiguration'
local WFFiniteStateMachine = require 'WFFiniteStateMachine'

	-- pixel provider
local PixelPathProviderStateMachine = require 'PixelPathProviderStateMachine'
local PixelProviderSharedBlobCache = require 'PixelProviderSharedBlobCache'
local WFCatRenderSize = require 'WFCatRenderSize'
local NmbLocalRenditionDigest = require 'NmbLocalRenditionDigest'

--------------------------------------------------------------------------------

local logger = require 'WFLogger'( 'PixelProviderChooseRenditionStrategyTask' )

--------------------------------------------------------------------------------

local PixelCacheCheckStates = 
	PixelPathProviderStateMachine.PixelCacheCheckStates

--==============================================================================
-- choose pixel strategy 
--==============================================================================

local SaveRenderStates = PixelPathProviderStateMachine.SaveRenderStates

--------------------------------------------------------------------------------

local needSaveRenderStates = 
			{ WFFiniteStateMachine.makeRetryableStates(
					PixelPathProviderStateMachine.SaveRenderRetryCount, 
					SaveRenderStates.NeedToSaveRender ) }

--------------------------------------------------------------------------------

local PixelProviderSharedBlobCache_getEquivalentRenditionRequests =
		PixelProviderSharedBlobCache.getEquivalentRenditionRequests

--------------------------------------------------------------------------------

local function canJoinPendingSave( scheduler, activeRenditionInfoCache,
									localAssetId, renderSize, digest )

	local currentPendingSaveRequests = 
			PixelProviderSharedBlobCache_getEquivalentRenditionRequests( 
				scheduler, activeRenditionInfoCache, localAssetId,
				renderSize, digest, unpack( needSaveRenderStates ) )

	return #currentPendingSaveRequests > 0

end

--------------------------------------------------------------------------------

local isSavingStates =
			{ WFFiniteStateMachine.makeRetryableStates(
					PixelPathProviderStateMachine.SaveRenderRetryCount, 
					SaveRenderStates.SavingRender ) }

--------------------------------------------------------------------------------

local function canJoinInProgressSave( scheduler, activeRenditionInfoCache,
										localAssetId, renderSize, digest )

	local currentInProgressSaveRequests = 
			PixelProviderSharedBlobCache_getEquivalentRenditionRequests( 
				scheduler, activeRenditionInfoCache, localAssetId,
				renderSize, digest, unpack( isSavingStates ) )

	return #currentInProgressSaveRequests > 0

end

--------------------------------------------------------------------------------

local canRenderWithProxy = 
		assert( WFCatRenderSize.canRenderWithProxy,
				"WFCatRenderSize.canRenderWithProxy no defined" )

--------------------------------------------------------------------------------

local function hasAtLeastOneLocalNegativeForRender( currOriginalPath, currProxyPath,
											neededLongSide )

	return currOriginalPath ~= nil or
			currProxyPath and canRenderWithProxy( neededLongSide )

end

--------------------------------------------------------------------------------

local function hasLocalProxySuitableForRender( currProxyPath, neededLongSide )

	return currProxyPath and canRenderWithProxy( neededLongSide )

end

--------------------------------------------------------------------------------

local neverDownload = not not WFConfiguration.PixelProvider.NeverDownload

--------------------------------------------------------------------------------

local function hasProxy( baseRenderInfo ) 
	
	local proxies = baseRenderInfo.proxies or {}
	
	return ( #proxies > 0 )
	
end

--------------------------------------------------------------------------------

local function hasOriginal( baseRenderInfo )

	return not not baseRenderInfo.master

end

--------------------------------------------------------------------------------

local function maybeGetTempDigestAndRenditionPathForMissingProfile( 
	scheduler, renditionPathBaseDir, pixStreamId, 
	localAssetId, baseRenderInfo, activeRenditionInfo )

	if baseRenderInfo.hasStandardProfile and baseRenderInfo.cameraProfileMissing then

		logger:warnf( "Render with digest for missing profile %s. %s: no profile",
						pixStreamId, localAssetId )

		logger:warnf( "Standard profile unavailable, include missing info %s", localAssetId )
		
		local temporaryDigest = 
				NmbLocalRenditionDigest.computeTemporaryDefaultForOrientation( 
					baseRenderInfo.ozAssetId or localAssetId, 
					activeRenditionInfo.devSettings, activeRenditionInfo.userOrientation,
					baseRenderInfo.cameraProfileDigest, baseRenderInfo.cameraProfileMissing,
					true )
		
		logger:warnf( "Actual digest:%s Temp Digest:%s", activeRenditionInfo.digest, temporaryDigest )
		
		local tempPath = require "WFCatPixelPathGenerator".pathForPreviewDigestAndLongEdge( 
			renditionPathBaseDir, baseRenderInfo.ozAssetId, 
			temporaryDigest, activeRenditionInfo.renderSize )
		
		logger:warnf( "Actual path:%s Temp path:%s", activeRenditionInfo.renditionPath, tempPath )
		
		return temporaryDigest, tempPath

	end

end

--------------------------------------------------------------------------------

local function maybeChooseNonRenderState( request, scheduler,
											baseRenderInfoCache,
											activeRenditionInfoCache )

	local pixStreamId, localAssetId = unpack( request )

	local activeRenditionInfo = activeRenditionInfoCache:peekCacheItem( 
										pixStreamId, localAssetId )
	
	if not activeRenditionInfo then

		logger:warnf( "Can't choose pixel strategy for %s.%s: no renderInfo",
						pixStreamId, localAssetId )

		-- force it to retrieve again
		return scheduler:_applyTransitionToRequest( 
							request, 'new_docstore_render_info' )

	end
	
	local wantedDigest = activeRenditionInfo.digest
	
	local renderSize = activeRenditionInfo.renderSize

	if canJoinPendingSave( scheduler, activeRenditionInfoCache,
								localAssetId, renderSize, wantedDigest ) then

		return scheduler:_applyTransitionToRequest( 
							request, 'has_pending_rendition_save' )

	end
	
	if canJoinInProgressSave( scheduler, activeRenditionInfoCache,
								localAssetId, renderSize, wantedDigest ) then

		return scheduler:_applyTransitionToRequest( 
							request, 'has_active_rendition_save' )

	end

	local baseRenderInfo = baseRenderInfoCache:peekCacheItem( localAssetId )
	
	if not baseRenderInfo then

		logger:warnf( "Can't choose pixel strategy for %s.%s: no renderInfo",
						pixStreamId, localAssetId )

		-- force it to retrieve again
		return scheduler:_applyTransitionToRequest( 
							request, 'new_docstore_render_info' )

	end

end

--------------------------------------------------------------------------------

local function maybeChooseLocalRender( request, scheduler, baseRenderInfo,
										renderSize, renditionPathBaseDir, pixStreamId, 
										localAssetId, activeRenditionInfo, 
										baseRenderInfoCache, activeRenditionInfoCache )

	local currOriginalPath, currProxyPath, originalCannotBeRendered = 
			baseRenderInfo.originalPath, baseRenderInfo.proxyPath, baseRenderInfo.originalCannotBeRendered
			
	if originalCannotBeRendered then
		currOriginalPath = nil
	end

	local checkNegativeTransition =
		( hasAtLeastOneLocalNegativeForRender( currOriginalPath, currProxyPath, 
												renderSize ) and 
		  'confirm_any_local_negative_for_render' ) or
		( hasLocalProxySuitableForRender( currProxyPath, renderSize ) and 
		  'confirm_local_proxy_for_render' )

	if not checkNegativeTransition then
		return
	end

	local maybeTempDigest, maybePath =
		maybeGetTempDigestAndRenditionPathForMissingProfile( scheduler, 
			renditionPathBaseDir,
			pixStreamId, 
			localAssetId, 
			baseRenderInfo, 
			activeRenditionInfo )

	if maybeTempDigest then

		baseRenderInfo.digest = maybeTempDigest
			
		baseRenderInfoCache:pushCacheItem( localAssetId, baseRenderInfo )

		activeRenditionInfo.digest = maybeTempDigest
		activeRenditionInfo.renditionPath = maybePath
			
		activeRenditionInfoCache:pushCacheItem( pixStreamId, localAssetId,
										activeRenditionInfo )
	end

	return scheduler:_applyTransitionToRequest( 
								request, checkNegativeTransition )

end

--------------------------------------------------------------------------------

local function haveSuitableDownloadableRenditon( baseRenderInfoCache, localAssetId, renderSize )

	if neverDownload then

		return false

	end

	local maybeDownloadableRenditionSizeName = 
			PixelProviderSharedBlobCache.maybeGetDownloadableRenditionSizeName(
				baseRenderInfoCache, localAssetId, renderSize )
			
	return maybeDownloadableRenditionSizeName ~= nil
	
end

--------------------------------------------------------------------------------

local function chooseRenditionStrategy( request, scheduler, renditionPathBaseDir, baseRenderInfoCache,
											activeRenditionInfoCache,
											currRenderSettingsCache,
											updatedDevelopSettingsAssetIds )

	local maybeNonRenderState = 
			maybeChooseNonRenderState( 
				request, scheduler, baseRenderInfoCache,
				activeRenditionInfoCache )

	if maybeNonRenderState then

		return maybeNonRenderState

	end

	local pixStreamId, localAssetId = unpack( request )

	local activeRenditionInfo = activeRenditionInfoCache:peekCacheItem( 
										pixStreamId, localAssetId )
	
	local renderSize = activeRenditionInfo.renderSize
	
	local baseRenderInfo = baseRenderInfoCache:peekCacheItem( localAssetId )
	local hasCustomProfileLocally = baseRenderInfo.cameraProfileDigest
				and not baseRenderInfo.cameraProfileMissing
				and not baseRenderInfo.hasStandardProfile

	local canDownloadRendition =
			haveSuitableDownloadableRenditon( baseRenderInfoCache, localAssetId,
											renderSize ) 
			and not updatedDevelopSettingsAssetIds[ localAssetId ]
			and not hasCustomProfileLocally
			
	
	local preferDownloadsOverRenders = 
			scheduler:shouldPreferDownloadsForNewRenditions()
	
	if canDownloadRendition and preferDownloadsOverRenders then
	
		return scheduler:_applyTransitionToRequest( 
				request, 'try_download_rendition' )

	end

	-- Since we're unable to download a rendition from Oz we will instead render.
	-- We first need to download/read any external XMP settings in order to render.

	local _, _, xmpSettings = 
			unpack( currRenderSettingsCache:peekCacheItem( localAssetId ) or {} )

	if not xmpSettings then
		
		if canDownloadRendition then

			return scheduler:_applyTransitionToRequest( 
					request, 'try_download_rendition' )

		end
		
		currRenderSettingsCache:popCacheItem( localAssetId )
			-- clear out any previous render settings while we try to fetch the new 
			-- external settings
		
		return scheduler:_applyTransitionToRequest( request, 'need_develop_xmp' )
	
	end

	local maybeLocalRenderState = 
			maybeChooseLocalRender( request, scheduler, baseRenderInfo,
										renderSize, renditionPathBaseDir, pixStreamId, 
										localAssetId, activeRenditionInfo, 
										baseRenderInfoCache, activeRenditionInfoCache )

	if maybeLocalRenderState then

		return maybeLocalRenderState

	end
	
	-- we can't render
	
	-- TODO [matthewj 2016-03-31]: We need to be able to consider the PANO proxy case
	if canDownloadRendition then

		return scheduler:_applyTransitionToRequest( 
				request, 'try_download_rendition' )
	
	elseif canRenderWithProxy( renderSize ) and hasProxy( baseRenderInfo ) then
		
		return scheduler:_applyTransitionToRequest( 
							request, 'try_download_proxy' )
	
	elseif hasOriginal( baseRenderInfo ) then
		
		return scheduler:_applyTransitionToRequest( 
							request, 'try_download_original' )
											
	end

	return scheduler:_applyTransitionToRequest( request, 'no_original' )

end

--------------------------------------------------------------------------------

local function chooseRenderOnlyRenditionStrategy( 
					request, scheduler, renditionPathBaseDir, baseRenderInfoCache,
					activeRenditionInfoCache, currRenderSettingsCache )

	local maybeNonRenderState = 
			maybeChooseNonRenderState( 
				request, scheduler, baseRenderInfoCache,
				activeRenditionInfoCache )

	if maybeNonRenderState then

		return maybeNonRenderState

	end

	local pixStreamId, localAssetId = unpack( request )

	-- We first need to download/read any external XMP settings in order to render.

	local _, _, xmpSettings = 
			unpack( currRenderSettingsCache:peekCacheItem( localAssetId ) or {} )

	if not xmpSettings then
		
		currRenderSettingsCache:popCacheItem( localAssetId )
			-- clear out any previous render settings while we try to fetch the new 
			-- external settings
		
		return scheduler:_applyTransitionToRequest( request, 'need_develop_xmp' )
	
	end

	local activeRenditionInfo = activeRenditionInfoCache:peekCacheItem( 
										pixStreamId, localAssetId )
	
	local renderSize = activeRenditionInfo.renderSize

	local baseRenderInfo = baseRenderInfoCache:peekCacheItem( localAssetId )

	local maybeLocalRenderState = 
			maybeChooseLocalRender( request, scheduler, baseRenderInfo,
										renderSize, renditionPathBaseDir, pixStreamId, 
										localAssetId, activeRenditionInfo, 
										baseRenderInfoCache, activeRenditionInfoCache )

	if maybeLocalRenderState then

		return maybeLocalRenderState

	end

	return scheduler:_applyTransitionToRequest( request, 'no_original' )

end

--------------------------------------------------------------------------------

local function wrapInSendTestData( f )

	return function( _, scheduler, ... )

		local ret = f( _, scheduler, ... )

		scheduler:_sendTestData( "chooseRenditionStrategy", ret )

		return ret

	end

end

--------------------------------------------------------------------------------

chooseRenditionStrategy = wrapInSendTestData( chooseRenditionStrategy )

--------------------------------------------------------------------------------

local function processRequests( scheduler, taskState, baseRenderInfoCache, 
									activeRenditionInfoCache,
									currRenderSettingsCache,
									updatedDevelopSettingsAssetIds )

	-- update all of the ChooseRenditionStrategy requests
	scheduler:_mapAllRequestsWithState( 
		PixelCacheCheckStates.ChooseRenditionStrategy,
			chooseRenditionStrategy, 
				scheduler, taskState.renditionPathBaseDir, 
				baseRenderInfoCache, activeRenditionInfoCache,
				currRenderSettingsCache, updatedDevelopSettingsAssetIds )

	-- update all of the ChooseRenderOnlyRenditionStrategy requests
	scheduler:_mapAllRequestsWithState( 
		PixelCacheCheckStates.ChooseRenderOnlyRenditionStrategy,
			chooseRenderOnlyRenditionStrategy, 
				scheduler, taskState.renditionPathBaseDir, 
				baseRenderInfoCache, activeRenditionInfoCache,
				currRenderSettingsCache )

end
--==============================================================================
-- setupTask
--==============================================================================

local function setupTask( renditionPathBaseDir )
	
	return { 

		renditionPathBaseDir = renditionPathBaseDir,
			
	}

end

--==============================================================================

return {

	setupTask = setupTask,
	processRequests = processRequests,

}
