--[[----------------------------------------------------------------------------

WFCatCollaboration.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
local Rx = require 'Rx'
local Observable = Rx.Observable

local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFTableMerge = require 'WFTableMerge'
local WFRxScheduler = require 'WFRxScheduler'
local WFRxThrottledBatchedBulkApplyOverlay = require 'WFRxThrottledBatchedBulkApplyOverlay'
local WFRxThrottledBatchedBulkDocValue = require 'WFRxThrottledBatchedBulkDocValue'

local WFCatCollaborationConstants = require 'WFCatCollaborationConstants'
local WFCatUtils = require 'WFCatUtils'
local WFLogger = require 'WFLogger'

local WFRxDocValueWatcher = require 'WFRxDocValueWatcher'

--------------------------------------------------------------------------------

local check = WFArguments.check
local derefOpt = KSTableUtils.derefOpt
local logger = WFLogger( 'WFCatCollaboration' )

--==============================================================================

-- @mark WFCatCollaboration namespace

--------------------------------------------------------------------------------

local WFCatCollaboration = {}

--------------------------------------------------------------------------------

local function intToBool( zeroOrOne )
	return zeroOrOne ~= 0
end

--------------------------------------------------------------------------------

function WFCatCollaboration.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	-- table space

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space( created, updateTimeStamp, private, download, allowInviteRequests )]],
		docValueFilter = function( docValue )
			return docValue.subtype == 'shared' and docValue.type == 'space'
		end,
		hasOneRowPerDocument = true,
		
		map = function( docValue, emit )

			if docValue and not docValue._deleted
			  and docValue.type == 'space' and docValue.subtype == 'shared'
			then

				local localOnly = docValue._localOnly or {}
				
				local isPrivate = docValue.private
				local canDownload = docValue.download
				local allowInviteRequests = docValue.allowInviteRequests
				local updateTimeStamp = docValue.userUpdated or docValue.updated
				local created = docValue.created or localOnly.created or
								derefOpt( docValue, '_serverProvidedMetadata', 'created' )
								
				emit( created, updateTimeStamp, isPrivate and 1 or 0, canDownload and 1 or 0, allowInviteRequests and 1 or 0 )
					
			end
								
		end,

	}
	
	indexStoreSchema:addIndex [[CREATE INDEX space_private_index ON space( docId, private )]]

	-- table space_album

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space_album( spaceId, albumId, created, updateTimeStamp )]],
		docValueFilter = function( docValue )
			return docValue.type == 'space_album'
		end,
		map = function( docValue, emit )

			if docValue and not docValue._deleted
			  and docValue.type == 'space_album'
			then

				local localOnly = docValue._localOnly or {}
				
				local albumId = docValue.album and docValue.album.id
				local spaceId = docValue.space and docValue.space.id
				local updateTimeStamp = docValue.userUpdated or docValue.updated
				local created = docValue.created or localOnly.created or
								derefOpt( docValue, '_serverProvidedMetadata', 'created' )
				
				if spaceId and albumId then
					
					emit( spaceId, albumId, created, updateTimeStamp )
					
				end
					
			end
						
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX space_album_index ON space_album( spaceId, albumId )]]
	indexStoreSchema:addIndex [[CREATE INDEX space_album_album ON space_album( albumId )]]
	
	
	-- table space_invite

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space_invite( subtype, spaceId, userId, email, status, role, created, updated, space_user )]],
		docValueFilter = function( docValue )
			return docValue.type == 'spaceinvite'
				and ( docValue.subtype == 'request' or docValue.subtype == 'single_use' )
		end,
		map = function( docValue, emit )

			if docValue and not docValue._deleted then

				local subtype = derefOpt( docValue, 'subtype' )
				local spaceId = derefOpt( docValue, 'space', 'id' )
				local userId = derefOpt( docValue, 'user', 'id' )
				local created = derefOpt( docValue, 'userCreated' ) or derefOpt( docValue, '_serverProvidedMetadata', 'created' )
				local updated = derefOpt( docValue, 'userUpdated' ) or derefOpt( docValue, '_serverProvidedMetadata', 'updated' )
				local email = derefOpt( docValue, 'email' )
				local space_user = derefOpt( docValue, 'space_user', 'id' )
				local status = derefOpt( docValue, 'status' )
				local role = derefOpt( docValue, 'role' )
				
				if spaceId and subtype and created then
					
					emit( subtype, spaceId, userId, email, status, role, created, updated, space_user )
					
				end
					
			end
			
		end,

		useLocalId = true,
		hasOneRowPerDocument = true,
	}

	indexStoreSchema:addIndex [[CREATE INDEX space_invite_index ON space_invite( spaceId, subtype )]]
	indexStoreSchema:addIndex [[CREATE INDEX space_invite_user ON space_invite( userId )]]
	

	-- table space_user

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space_user( spaceId, userId, status, role, created, updated )]],
		docValueFilter = function( docValue )
			return docValue.type == 'space_user'
		end,
		map = function( docValue, emit )

			if docValue and not docValue._deleted then

				local spaceId = derefOpt( docValue, 'space', 'id' )
				local userId = derefOpt( docValue, 'user', 'id' )
				local status = derefOpt( docValue, 'status' )
				local role = derefOpt( docValue, 'role' )
				local created = derefOpt( docValue, 'userCreated' ) or derefOpt( docValue, '_serverProvidedMetadata', 'created' )
				local updated = derefOpt( docValue, 'userUpdated' ) or derefOpt( docValue, '_serverProvidedMetadata', 'updated' )
				
				if spaceId and userId then
					
					emit( spaceId, userId, status, role, created, updated )
					
				end
					
			end
						
		end,

		useLocalId = true,
		hasOneRowPerDocument = true,
	}

	indexStoreSchema:addIndex [[CREATE INDEX space_user_index ON space_user( spaceId )]]
	indexStoreSchema:addIndex [[CREATE INDEX space_user_user ON space_user( userId )]]
	
end

--------------------------------------------------------------------------------

local function getSyncTerms()

	return {
		fetchSql = [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]],
		isShared = [[SELECT
						space_album.docId, spaceId FROM space_album
					INNER JOIN
						space
					ON
						space_album.spaceId == space.docId
					WHERE
						albumId == ? AND space.private == 0]],
		docId = 'albumId',
		space_doc_id = 'space_album_id',
		space_doc = 'space_album',
	}
end

--------------------------------------------------------------------------------

local function maybeGetFullAlbumId( catalog, maybeLocalAlbumId )

	if type( maybeLocalAlbumId ) == 'string' then
		return Observable.just( maybeLocalAlbumId )
	end
	
	return catalog:docStore():localIdToDocId( maybeLocalAlbumId )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.inviteByEmail( catalog, maybeLocalAlbumId, email, optAccessLevel )

	local accessLevel = optAccessLevel or WFCatCollaborationConstants.kRoles.viewer

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )

			local syncTerms = getSyncTerms()

			return catalog:indexStore():queryOnce( {
					sql = syncTerms.fetchSql,
					quiesceUntilUpToDate = true,
					fullAlbumId,
				} )
				:flatMap( function( idRows )

					local spaceId = derefOpt( idRows, 1, 2 )

					if spaceId then
					
						local id = KSIdGenerator.uuidInLowercaseHex()
					
						local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( id )

						local spaceinviteDocument = {
							type = 'spaceinvite',
							subtype = 'single_use',
							email = email,
							status = WFCatCollaborationConstants.kSingleUseInviteStatuses.open,
							role = accessLevel,
							space = {
								id = spaceId,
							},
						}
	
						WFTableMerge.applyOverlay( spaceinviteDocument, baseDocValue )
	
						return catalog:docStore():setDocValue( spaceinviteDocument )
					end
				end )
		end )	
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getCollaborators( catalog, maybeLocalAlbumId )

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return catalog:indexStore():query( {
				sql = [[ SELECT
							space_user.docId, space_user.userId, space_user.status, space_user.role, space_user.created, space_user.updated, feedback_user.fullName, feedback_user.email
						FROM
							space_user
						INNER JOIN
							space_album
						ON
							space_user.spaceId == space_album.spaceId
						INNER JOIN
							feedback_user
						ON
							space_user.userId == feedback_user.userId
						WHERE
							space_album.albumId ==  ? ]],
				quiesceUntilUpToDate = true,
				calm = 2,
				
				fullAlbumId,
			} )
				:map( function( rows )
					local collaborators = {}
					
					for _, row in ipairs( rows ) do
						collaborators[ #collaborators + 1 ] = {
							id = row[ 1 ],
							userId = row[ 2 ],
							status = row[ 3 ],
							role = row[ 4 ],
							created = row[ 5 ],
							updated = row[ 6 ],
							fullName = row[ 7 ],
							email = row[ 8 ],
						}
					end
					
					return collaborators
				end )
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getPendingRequests( catalog, maybeLocalAlbumId )

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return catalog:indexStore():query( {
				sql = [[ SELECT
							space_invite.docId, space_invite.userId, space_invite.created, space_invite.updated, space_invite.status, space_invite.role, space_invite.space_user, feedback_user.fullName, feedback_user.email
						FROM
							space_invite
						INNER JOIN
							space_album
						ON
							space_invite.spaceId == space_album.spaceId
						INNER JOIN
							feedback_user
						ON
							space_invite.userId == feedback_user.userId
						WHERE
							space_invite.subtype == 'request' AND space_album.albumId == ? ]],
				quiesceUntilUpToDate = true,
				calm = 2,

				fullAlbumId,
			} )
				:map( function( rows )
					local requests = {}
					
					for _, row in ipairs( rows ) do
						requests[ #requests + 1 ] = {
							id = row[ 1 ],
							userId = row[ 2 ],
							created = row[ 3 ],
							updated = row[ 4 ],
							status = row[ 5 ],
							role = row[ 6 ],
							pendingApproval = row[ 7 ] == nil,
							fullName = row[ 8 ],
							email = row[ 9 ],
						}
					end
					
					return requests
				end )
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getInvitations( catalog, maybeLocalAlbumId )

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return catalog:indexStore():query( {
				sql = [[ SELECT
							space_invite.docId, space_invite.email, space_invite.created, space_invite.updated, space_invite.status, space_invite.role
						FROM
							space_invite
						INNER JOIN
							space_album
						ON
							space_invite.spaceId == space_album.spaceId
						WHERE
							space_invite.subtype == 'single_use' AND space_album.albumId == ? ]],
				quiesceUntilUpToDate = true,
				calm = 2,

				fullAlbumId,
			} )
				:map( function( rows )
					local invitations = {}
					
					for _, row in ipairs( rows ) do
						invitations[ #invitations + 1 ] = {
							id = row[ 1 ],
							email = row[ 2 ],
							created = row[ 3 ],
							updated = row[ 4 ],
							status = row[ 5 ],
							role = row[ 6 ],
						}
					end
					
					return invitations
				end )
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.updateCollborators( catalog, updates )
	
	local docStore = catalog:docStore()
	
	local docIds = {}
	for _, update in ipairs( updates ) do
		docIds[ #docIds + 1 ] = update.id
	end
	
	return WFRxThrottledBatchedBulkDocValue( docStore, docIds, 100, 10 )
		:flatMap( function( docValues )
		
			local updateOverlays = {}
			
			for i, update in ipairs( updates ) do
			
				local docValue = docValues[ i ]
				local docType = docValue.type
				local subType = docValue.subtype
				
				local optDelete = update.delete
				local optStatus = update.status
				local optRole = update.role
			
				local overlay
				
				if optDelete then
					overlay = {
						_id = docValue._id,
						_deleted = true,
					}
				else
					if docType ~= 'space_user' and docType ~= 'spaceinvite' then
						return Observable.throw( 'Not a collaborator document' )
					end
					
					if optRole and not WFCatCollaborationConstants.kRoles[ optRole ] then
						return Observable.throw( 'Unknown role' )
					end
					
					if optStatus ~= nil and (
						( docType == 'space_user' and not WFCatCollaborationConstants.kSpaceUserStatuses[ optStatus ] ) or
						( docType == 'spaceinvite' and subType == 'request' and not WFCatCollaborationConstants.kRequestInviteStatuses[ optStatus ] ) or
						( docType == 'spaceinvite' and subType == 'single_use' and not WFCatCollaborationConstants.kSingleUseInviteStatuses[ optStatus ] )
					) then
						return Observable.throw( 'Unknown status' )
					end
						
					overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( docValue._id )
					
					overlay.status = optStatus
					overlay.role = optRole
				end
				
				updateOverlays[ i ] = overlay
			
			end
			
			return WFRxThrottledBatchedBulkApplyOverlay( docStore, updateOverlays, 100, 10 )
		
		end )	
end

--------------------------------------------------------------------------------

local function removeAllCollaborators( catalog, albumId )
	
	local allCollaboratorsObservable = WFCatCollaboration.getCollaborators( catalog, albumId )
		:take( 1 )

	local allPendingRequestsObservable = WFCatCollaboration.getPendingRequests( catalog, albumId )
		:take( 1 )

	local allInvitationsObservable = WFCatCollaboration.getInvitations( catalog, albumId )
		:take( 1 )
	
	return Observable.combineLatest( allCollaboratorsObservable, allPendingRequestsObservable, allInvitationsObservable )
		:flatMap( function( allCollaborators, allPendingRequests, allInvitations )
			
			local allOverlays = {}
			
			for _, collaborator in ipairs( allCollaborators ) do
				allOverlays[ #allOverlays + 1 ] = {
					_id = collaborator.id,
					_deleted = true,
				}
			end
			
			for _, request in ipairs( allPendingRequests ) do
				allOverlays[ #allOverlays + 1 ] =  {
					_id = request.id,
					_deleted = true,
				}
			end
			
			for _, invitation in ipairs( allInvitations ) do
				allOverlays[ #allOverlays + 1 ] =  {
					_id = invitation.id,
					_deleted = true,
				}
			end
			
			if #allOverlays == 0 then
				return Observable.just( true )
			end
			
			local docStore = catalog:docStore()
			
			return WFRxThrottledBatchedBulkApplyOverlay( docStore, allOverlays, 100, 10 )
			
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getAllSharedAlbums( catalog ) 
	
	return catalog:indexStore():query {
			sql = [[ SELECT albumId, spaceId, private, allowInviteRequests from space_album JOIN space ON space_album.spaceId == space.docId ]],
			calm = 2,
		}
		:map( function( sharedAlbums )
			local results = {}
			for _, sharedAlbum in ipairs( sharedAlbums ) do 
				local result = {
					albumId = sharedAlbum[ 1 ],
					spaceId = sharedAlbum[ 2 ],
					isPrivate = intToBool( sharedAlbum[ 3 ] ),
					allowInviteRequests = intToBool( sharedAlbum[ 4 ] ),
				}
				table.insert( results, result )
			end
			return results
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getSharingStatusByAlbumId( catalog )

	local indexStore = catalog:indexStore()

	local function getCollaborationEntries( tableName, subType )
	
		local sqlStr = string.format( "SELECT spaceId, status FROM %s", tableName )
		
		if subType then
			sqlStr = sqlStr .. string.format( " WHERE subType == '%s'", subType )
		end
	
		return indexStore:query( {
			sql = sqlStr,
			quiesceUntilUpToDate = true,
			calm = 2,
		} )
			:map( function( rows )
				
				local ret = {}
				for _, row in ipairs( rows ) do
					ret[ #ret + 1 ] = {
						space = row[ 1 ],
						status = row[ 2 ],
					}
				end
				
				return ret
				
			end )
			
	end
	
	local function groupBySpace( entriesObservable, filterFn )
		
		return entriesObservable
			:map( function( entries )
			
				local entriesBySpace = {}
				
				for _, entry in ipairs( entries ) do
					if filterFn( entry ) then
						local entriesForSpace = entriesBySpace[ entry.space ] or {}
						entriesForSpace[ #entriesForSpace + 1 ] = entry
						entriesBySpace[ entry.space ] = entriesForSpace
					end
				end
				
				return entriesBySpace
			
			end )
	end
	
	
	local collaboratorsBySpaceObservable = groupBySpace( getCollaborationEntries( 'space_user' ), function( entry )
		return entry.status == WFCatCollaborationConstants.kSpaceUserStatuses.accepted
	end )
	
	local requestsBySpaceObservable = groupBySpace( getCollaborationEntries( 'space_invite', 'request' ), function( entry )
		return entry.status == WFCatCollaborationConstants.kRequestInviteStatuses.open
	end )
	
	local invitationsBySpaceObservable = groupBySpace( getCollaborationEntries( 'space_invite', 'single_use' ), function( entry )
		return entry.status == WFCatCollaborationConstants.kSingleUseInviteStatuses.open
	end )
	
	return Observable.combineLatest(
		WFCatCollaboration.getAllSharedAlbums( catalog ),
		collaboratorsBySpaceObservable,
		requestsBySpaceObservable,
		invitationsBySpaceObservable
	)
		:map( function( sharedAlbums, collaboratorsBySpace, requestsBySpace, invitationsBySpace )
			
			local retSharesByAlbumId = {}
			
			for _, sharedAlbum in ipairs( sharedAlbums ) do
			
				local albumId = sharedAlbum.albumId
				local spaceId = sharedAlbum.spaceId
				
				local hasCollaborators = collaboratorsBySpace[ spaceId ] and next( collaboratorsBySpace[ spaceId ] ) ~= nil
				local hasPendingRequests = requestsBySpace[ spaceId ] and next( requestsBySpace[ spaceId ] ) ~= nil
				local hasInvitations = invitationsBySpace[ spaceId ] and next( invitationsBySpace[ spaceId ] ) ~= nil

				local isSharedFromSettings = not sharedAlbum.isPrivate or sharedAlbum.allowInviteRequests
				
				retSharesByAlbumId[ albumId ] = ( isSharedFromSettings or hasCollaborators or hasPendingRequests or hasInvitations ) and true or false
			end
			
			return retSharesByAlbumId
			
		end )
end

--------------------------------------------------------------------------------

local function setAttribute( dst, src, attribute, fallbackToDefault, default )
	if src[ attribute ] ~= nil then
		dst[ attribute ] = src[ attribute ]
	elseif fallbackToDefault then
		dst[ attribute ] = default
	end
end

--------------------------------------------------------------------------------

local function setSpaceAttributes( space, optAttributes, optApplyDefaults )
	local fallbackToDefault = optApplyDefaults
	if fallbackToDefault == nil then
		fallbackToDefault = true
	end
	
	local attr = optAttributes or {}
	setAttribute( space, attr, 'download', fallbackToDefault, false )
	setAttribute( space, attr, 'metadata', fallbackToDefault, false )
	setAttribute( space, attr, 'location', fallbackToDefault, false )
	setAttribute( space, attr, 'private', fallbackToDefault, true )
	setAttribute( space, attr, 'comments_disabled', fallbackToDefault, false )
	setAttribute( space, attr, 'favorites_disabled', fallbackToDefault, false )
	setAttribute( space, attr, 'allowInviteRequests', fallbackToDefault, true )
	space.presentationFilters = true
	return space
end

--------------------------------------------------------------------------------

local function setAlbumAttributes( albumOverlay, optAttributes, optApplyDefaults )
	local fallbackToDefault = optApplyDefaults
	if fallbackToDefault == nil then
		fallbackToDefault = true
	end
	
	local attr = optAttributes or {}

    local story = albumOverlay.story or {}
	setAttribute( story, attr, 'author', false )
	setAttribute( story, attr, 'title', false )
	setAttribute( story, attr, 'description', false )
	albumOverlay.story = story
	
	local presentation = albumOverlay.presentation or {}
	
	local options = presentation.options or {}
	setAttribute( options, attr, 'showTitle', fallbackToDefault, true )
	setAttribute( options, attr, 'showAuthor', fallbackToDefault, true )
	setAttribute( options, attr, 'showCaptions', fallbackToDefault, false )
	setAttribute( options, attr, 'showCover', fallbackToDefault, false )
	setAttribute( options, attr, 'cover', false )
	presentation.options = options
	
	local explore = presentation.explore or {}
	setAttribute( explore, explore, 'version', fallbackToDefault, 1 )
	setAttribute( explore, attr, 'themeId', fallbackToDefault, 'Gallery' )
	setAttribute( explore, attr, 'styleId', fallbackToDefault, 'Dark' )
	presentation.explore = explore
	
	local filters = presentation.filters or {}
	setAttribute( filters, attr, 'pickedOnly', fallbackToDefault, false )
	setAttribute( filters, attr, 'ratingAtLeast', fallbackToDefault, 0 )
	presentation.filters = filters
	
	albumOverlay.presentation = presentation
	return albumOverlay
end

--------------------------------------------------------------------------------

local function getAlbumAttributes( attributes, optAlbumDoc )

	local doc = optAlbumDoc
	if attributes and doc then
		local story = derefOpt( doc, 'story'  ) or {}
		setAttribute( attributes, story, 'author', false )
		setAttribute( attributes, story, 'title', false )
		setAttribute( attributes, story, 'description', false )
	
		local options = derefOpt( doc, 'presentation', 'options' ) or {}
		setAttribute( attributes, options, 'showTitle', true, true )
		setAttribute( attributes, options, 'showAuthor', true, true )
		setAttribute( attributes, options, 'showCaptions', true, false )
		setAttribute( attributes, options, 'showCover', true, false )
		setAttribute( attributes, options, 'cover', false )
		
		local explore = derefOpt( doc, 'presentation', 'explore' ) or {}
		setAttribute( attributes, explore, 'themeId', true, 'Gallery' )
		setAttribute( attributes, explore, 'styleId', true, 'Dark' )
		
		local filters = derefOpt( doc, 'presentation', 'filters' ) or {}
		setAttribute( attributes, filters, 'pickedOnly', true, false )
		setAttribute( attributes, filters, 'ratingAtLeast', true, 0 )
    end 
	return attributes
end

--------------------------------------------------------------------------------

local function createSpaceDocument( spaceId, optAttributes )

	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceId )

	local spaceDocument = {
		type = 'space',
		subtype = "shared",
	}
	
	WFTableMerge.applyOverlay( spaceDocument, baseDocValue )
	
	return setSpaceAttributes ( spaceDocument, optAttributes )
end

--------------------------------------------------------------------------------

local function createSpaceAlbumDocument( spaceAlbumId, spaceId, fullDocId )
	
	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceAlbumId )
	
	local spaceAlbumDoc = {
		type = 'space_album',
		space = {
			id = spaceId,
		},
		album = {
			id = fullDocId,
		}
	}
	
	WFTableMerge.applyOverlay( spaceAlbumDoc, baseDocValue )
	
	return spaceAlbumDoc

end

--------------------------------------------------------------------------------

local function createShareWithOz( catalog, fullDocId, optAttributes )
	-- if we don't already have a URL, request it directly from oz
	local KSJsonEncoder = require "KSJsonEncoder"
	local KSJsonDecoder = require "KSJsonDecoder"
	local spaceId = KSIdGenerator.uuidInLowercaseHex()
	local spaceAlbumId = KSIdGenerator.uuidInLowercaseHex()
	local payload = setSpaceAttributes( {} , optAttributes )
	local data = KSJsonEncoder.encodeOzUploadJson(
		{
			payload= payload,
			subtype = "shared",
			resources = {
				{
					id = fullDocId,
					type = "album",
					space_album_id = spaceAlbumId,
				}
			}
		}
	)
	
	--  create space locally when online call fails
	local function createSpaceLocally()
			-- if we fail at calling oz, fall back to creating locally and syncing when we get back online ( URL will not update )
		logger:warnf( "Sharing album locally %s", fullDocId )
		local spaceDocument = createSpaceDocument( spaceId, optAttributes )

		local spaceAlbumDocument = createSpaceAlbumDocument( spaceAlbumId, spaceId, fullDocId )

		local newDocsObservable = catalog:docStore():bulkSetDocValue {
			spaceDocument,
			spaceAlbumDocument,
		}
		
		local albumOverlay = setAlbumAttributes( { _id = fullDocId, }, optAttributes )
		
		local updatedAlbumObservable = catalog:docStore():applyOverlay( albumOverlay )
		
		return Observable.combineLatest( newDocsObservable, updatedAlbumObservable )
			:map( function()
				
				local attributes = setSpaceAttributes ( { id = spaceDocument._id }, spaceDocument )

				getAlbumAttributes( attributes, albumOverlay )

				return attributes
			end )
	end

	local catalogSpec = catalog:ozCatalogSpec()
	if not catalogSpec then  -- if test environment doesn't support ozCatalogSpec
		return createSpaceLocally()
	end

	local requestUrl = string.format( '%s/spaces/%s', catalogSpec:catalogHref(), spaceId )
	return catalogSpec:put( requestUrl, data,
			{ 	requestMethod = 'PUT',
				requestHeaders = {
					{ header = 'Content-Type', value = 'application/json' },
				},
			} )
	:catch( createSpaceLocally )
	:delay( 0.5, WFRxScheduler.sharedScheduler() )
	:flatMap( function( data, headers )

		if headers and headers.status_code == "201" then

			return catalogSpec:get( requestUrl, { } )
			:catch( createSpaceLocally )
			:flatMap( function( data, headers )

				if headers and headers.status_code == "200" then
					local decodedSpaceJson = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
					local spaceDocument = require 'WFOzDocUtils'.remapBasicOzDocument( decodedSpaceJson, decodedSpaceJson.base )
					
					local href = derefOpt( spaceDocument, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )

					local spaceAlbumDocument = {
						type = "space_album",
						_id = spaceAlbumId,
						space = {
							id = spaceId,
						},
						album = {
							id = fullDocId,
						}
					}
					
					local newDocsObservable = catalog:docStore():bulkSetDocValue {
						label = 'oz',
						spaceDocument,
						spaceAlbumDocument,
					}
					
					local albumOverlay = setAlbumAttributes( { _id = fullDocId, }, optAttributes )
					
					local updatedAlbumObservable = catalog:docStore():applyOverlay( albumOverlay )
					
					return Observable.combineLatest( newDocsObservable, updatedAlbumObservable )
						:map( function()
						
							local attributes = setSpaceAttributes( { id = spaceId, url = href }, spaceDocument )
							
							getAlbumAttributes( attributes, albumOverlay )
							
							return attributes
						end )
				else
					return createSpaceLocally()
				end
			end)
		else
			return createSpaceLocally()
		end
	end)
end

--------------------------------------------------------------------------------

-- keeping this around because it is useful for testing and debugging
--local function getShareFromOz( catalog, spaceId )
--	-- if we don't already have a URL, request it directly from oz
--	local KSJsonDecoder = require "KSJsonDecoder" 
--	local catalogSpec = catalog:ozCatalogSpec()
--	local requestUrl = string.format( '%sspaces/%s', catalogSpec:baseHref(), spaceId )
--	return catalogSpec:get( requestUrl, { } )
--	:map( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty )
--	:map( function( docValue )
--		local href = derefOpt( docValue, 'links', '/rels/web_space', 'href'  )
--		return setSpaceAttributes ( { id = spaceId, url = href }, docValue )
--	end )
--end

--------------------------------------------------------------------------------

local function shareAlbum( catalog, docValue, spaceSubtype, optAttributes )
	
	-- The current sharing model for albums calls for a 1-1 mapping from an 'album' document to
	-- a 'space'. In order to share the document, a 'space' needs to be created and the document
	-- added to it.

	-- This function creates a space and adds the provided album to it by creating a
	-- 'space_album' document as appropriate. If a space already exists the existing space is used
	-- if optional attributes are given, they will be used in the new space or updated in the existing one
	-- This returns the attributes of the space
	
	local syncTerms = getSyncTerms()
	
	local fullDocId = docValue._id
		
	return catalog:indexStore():queryOnce {
			sql = syncTerms.fetchSql,
			quiesceUntilUpToDate = true,
			fullDocId
		}
		:flatMap( function( idRows )

			-- If there is a space for this album, get the ID for that space.
			local spaceId = derefOpt( idRows, 1, 2 )

			if spaceId then

				local spaceOverlay = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceId )
				setSpaceAttributes( spaceOverlay, optAttributes )
				
				local albumOverlay = setAlbumAttributes( { _id = fullDocId, }, optAttributes )
				
				return catalog:docStore():bulkApplyOverlay( { spaceOverlay, albumOverlay } )
					:flatMap( function()
						-- now read the updated space and return it
						return catalog:docStore():docValue( spaceId )
						:flatMap( function( spaceDocValue )
							local href = derefOpt( spaceDocValue, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )
							
							local attributes = setSpaceAttributes( { id = spaceId, url = href }, spaceDocValue )
							
							getAlbumAttributes( attributes, albumOverlay )
							
							return Observable.just( attributes )
						end )
					end )
			else  -- no space exists, so create one
				return createShareWithOz( catalog, fullDocId, optAttributes )
			end
		end )
end

--------------------------------------------------------------------------------

local function unshareAlbum( catalog, fullDocId )

	local syncTerms = getSyncTerms()
		
	return catalog:indexStore():queryOnce {
			sql = syncTerms.fetchSql,
			quiesceUntilUpToDate = true,
			fullDocId,
		}
		:flatMap( function( idRows )
		
			local spaceId = derefOpt( idRows, 1, 2 )
			
			if not spaceId then
				return 'error', "album is not shared"
			end
			
			local overlay = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceId )
			overlay.private = true
			overlay.allowInviteRequests = false
				
			return catalog:docStore():applyOverlay( overlay )
				:flatMap( function()
					return removeAllCollaborators( catalog, fullDocId )
				end )
					
		end )
end

--------------------------------------------------------------------------------

local function deleteAlbumShare( catalog, fullDocId )

	local syncTerms = getSyncTerms()
	
	return removeAllCollaborators( catalog, fullDocId )
		:flatMap( function()
	
			return catalog:indexStore():queryOnce {
					sql = syncTerms.fetchSql,
					quiesceUntilUpToDate = true,
					fullDocId,
				}
				:flatMap( function( idRows )
					
					-- SPAM [matthewj 2017-06-02]: Currently the sharing model is a
					-- 1-1 space to album relationship. Because of this when an 
					-- album is deleted, we know the space can be. As can all (only 1)
					-- space_album records. If this changes or additional
					-- space types are included in these shares this deletion needs
					-- to be rethought.
					local spaceResourceId = derefOpt( idRows, 1, 1 )
					local spaceId = derefOpt( idRows, 1, 2 )
					
					local deleteSpaceResource
					if spaceResourceId then
						deleteSpaceResource = catalog:docStore():bulkApplyOverlay( 
						{
							label = 'oz', -- don't send space resource deletes to Oz
							{ 
								_id = spaceResourceId, 
								_deleted = true,
							}
						} )
					else
						deleteSpaceResource = Observable.empty()
					end
					
					local deleteSpace
					if spaceId then
						deleteSpace = catalog:docStore():deleteDoc( spaceId )
					else
						deleteSpace = Observable.empty()
					end
					
					return Observable.combineLatest( deleteSpace, deleteSpaceResource )
					
				end )
				:takeLast( 0 ) 
				
		end )
end

--------------------------------------------------------------------------------

local function sharedAlbumUrl( catalog, fullDocId )

	local syncTerms = getSyncTerms()
	
	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			fullDocId
		}
		:pluck( 1, 2 )
		:flatMapLatest( function( spaceId )
			
			if spaceId then
				return catalog:docStore():ongoingDocValue( spaceId )
					:pluck( '_serverProvidedMetadata', 'links', 
							'/rels/web_space', 'href' )
			end
			
			return Observable.just( nil )
		
		end )
		:distinctUntilChanged()
end

--------------------------------------------------------------------------------

local function distinctTableCompare( _, tbl1, _, tbl2 )

	return KSTableUtils.deepEquals( tbl1, tbl2 )

end

--------------------------------------------------------------------------------

local function getShareAttributes( catalog, fullDocId )

	local syncTerms = getSyncTerms()

	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			fullDocId,
		}
		:pluck( 1, 2 ) -- Extract spaceId
		:filter( function( spaceId )
			return spaceId -- Only continue the Rx Chain if spaceId is not null.
		end )
		:take( 1 ) -- this take( 1 ) effectively closes down the index store query.
		:flatMap( function( spaceId )
		
			local spaceDocObservable = catalog:docStore():ongoingDocValue( spaceId )
			local albumDocObservable = catalog:docStore():ongoingDocValue( fullDocId )
			
			return Observable.combineLatest( spaceDocObservable, albumDocObservable )
				:flatMap( function( spaceDoc, maybeAlbumDoc )
				
					local href = derefOpt( spaceDoc, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )
					
					local attributes = setSpaceAttributes ( { id = spaceId, url = href }, spaceDoc )
					
					getAlbumAttributes( attributes, maybeAlbumDoc )
					
					return Observable.just( attributes )
					
				end )
				
		end )
		:distinctUntilChanged( nil, distinctTableCompare )
		
end

--------------------------------------------------------------------------------

local function setShareAttributes( catalog, docValue, attributes )

	local syncTerms = getSyncTerms()
	
	local fullDocId = docValue._id

	return catalog:indexStore():queryOnce {
			sql = syncTerms.fetchSql,
			quiesceUntilUpToDate = true,
			fullDocId,
		}
		:flatMap( function( idRows )
			local spaceId = derefOpt( idRows, 1, 2 )
			if not spaceId then
				return Observable.throw( "album is not shared" )
			end
			
			local spaceOverlay = setSpaceAttributes( { _id = spaceId, }, attributes, false )
			local albumOverlay = setAlbumAttributes( { _id = fullDocId, }, attributes, false )
			
			return catalog:docStore():bulkApplyOverlay( { spaceOverlay, albumOverlay } )
				:map( function()
					return true
				end )
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.sharedAlbumUrl( catalog, maybeLocalAlbumId )

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return sharedAlbumUrl( catalog, fullAlbumId )
		end )
end				

--------------------------------------------------------------------------------

function WFCatCollaboration.waitForAlbumSync( catalog, albumId )

	-- entrypoint for a end of sync test...
	
	return Observable.just( { albumId } )
	
		:call( WFRxDocValueWatcher.make, catalog:docStore() )
		
		:pluck( '_serverProvidedMetadata' )
		
		:filter( function( spm )
   			return not not spm
		end )
		
		:take( 1 )
		
		:flatMap( function()
			return Observable.just( albumId )
		end	)
end

--------------------------------------------------------------------------------

--[[!
	\brief deletes an album share. Generally only use this when the album itself is being 
		deleted. In most cases, this is not what you want as the share is
		permanently deleted. Unsharing an album removes the share from public view but
		keeps it's metadata (particularly it's URL) which allows share/unshare/share to
		be a non-destructive workflow. 
		
	\param albumDocId (string) The unique id of the album to remove shares for.
	
	\return <code>Observable< ></code> An Observable object that completes when 
		the operation is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.deleteSharedAlbum( catalog, albumDocId )
	
	return deleteAlbumShare( catalog, albumDocId )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief shares an album, creating the associated space and space_album documents.

	\param albumId (string|int) The unique or local ID of the album.
	\param optAttributes ( table ) optional attributes for creating the shared space

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.shareAlbum( catalog, albumId, optAttributes )

	
	return catalog:docStore():docValue( albumId )
		:flatMap( function( docValue )
			return shareAlbum( catalog, docValue, 'shared', optAttributes )
		end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Unshares an album, deleting the associated space and space_album documents.

	\param albumId (string|int) The unique or local ID of the album.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.unshareAlbum( catalog, maybeLocalAlbumId )
	
	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return unshareAlbum( catalog, fullAlbumId )
		end )

end

-------------------------------------------------------------------------------
--[[!
	\brief get shared album attributes.

	\param albumId (string|int) The unique or local ID of the album.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.getSharedAlbumAttributes( catalog, maybeLocalAlbumId )

	return maybeGetFullAlbumId( catalog, maybeLocalAlbumId )
		:flatMap( function( fullAlbumId )
			return getShareAttributes( catalog, fullAlbumId )
		end )
end

-------------------------------------------------------------------------------
--[[!
	\brief set shared album attributes.

	\param albumId (string|int) The unique or local ID of the album.
	\param attributes ( table ) updated attributes for the shared space

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.setSharedAlbumAttributes( catalog, albumId, attributes )

	return catalog:docStore():docValue( albumId )
		:flatMap( function( docValue )
			return setShareAttributes( catalog, docValue, attributes )
		end )
end
--------------------------------------------------------------------------------

return WFCatCollaboration
