--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local WFArguments = require 'WFArguments'
local WFConfiguration = require 'WFConfiguration'

	-- WF.oz_client.managed_catalog
local WFCatUser = require 'WFCatUser'
local WFCatSpec = require 'WFCatSpec'
local WFCatalog = require 'WFCatalog'
local WFCatMetadataPuller = require 'WFCatMetadataPuller'
local WFCatMetadataPusher = require 'WFCatMetadataPusher'
local WFCatIdleDownloader = require 'WFCatIdleDownloader'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local Observable = Rx.Observable

--------------------------------------------------------------------------------

local pusherDocumentBatchCount = WFConfiguration.Sync.pusherDocumentBatchCount
local pullerDocumentBatchCount = WFConfiguration.Sync.pullerDocumentBatchCount

--==============================================================================

-- @mark WFCatSyncClientImp namespace

--------------------------------------------------------------------------------

local syncClientImpMethods = {

	_user = false,
		-- (WFCatUser)
		
	_catSpec = false,
		-- (WFCatSpec)
		
	_catalog = false,
		-- (WFCatalog)
		
	_puller = false,
		-- (WFCatMetadataPuller)
		
	_pusher = false,
		-- (WFCatMetadataPusher)
		
	_checkOzForUpdatesSubscription = false,
		-- (Rx Disposable)
		
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

local function albumAssetsForFilteredAssetsQueryFn( catalog, assetIds )
	
	return catalog:indexStore():queryOnce {
				sql = [[SELECT album_asset_v2.docId, album_asset_v2.assetId FROM TEMP_ids JOIN album_asset_v2 ON 
							TEMP_ids.docId = album_asset_v2.assetId]],
				tempIds = assetIds,
			}
end

--------------------------------------------------------------------------------

function syncClientImpMethods:__init( options )

	check( 'WFCatSyncClientImp', 'table' )
	
	local serverBase = assert( options.serverBase )
	local userId = assert( options.userId )
	local userInfoCacheFolder = assert( options.userInfoCacheFolder )
	local catalogId = assert( options.catalogId )
	local rootCatalogPath = assert( options.rootCatalogPath )
	
	local pullerTaskDelays = options.metadataPullerArgs
	
	-- Set up sync infrastructure
	
	local user = WFCatUser( serverBase, userId, userInfoCacheFolder )
	
	local catSpec = WFCatSpec( user, options.catalogDoc or catalogId )
	
	local catArgs = {
		docStorePath = assert( options.docStorePath ),
		indexSpecScript = assert( options.indexSpecScript ),
	}
	
	local wfCatalog = WFCatalog( catSpec, rootCatalogPath, catArgs )
	
	-- Metadata Sync
	
	local puller = WFCatMetadataPuller( 
			wfCatalog, pullerTaskDelays, nil, nil, pullerDocumentBatchCount )
	
	local pusherOptions = {
		delayBeforeSendingUpdates = 2,
		documentBatchCount = pusherDocumentBatchCount or 100,
			-- Prevents the pusher from being overly aggressive at sending updates.
		albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn
	}
	
	local pusher = WFCatMetadataPusher( wfCatalog, pusherOptions )
		
	self:setOwned {
		_user = user,
		_catalog = wfCatalog,
		_puller = puller,
		_pusher = pusher,
	}
	
	
	-- WF-256: Reset MetadataPuller polling frequency after sending data to Oz.
	
	self._checkOzForUpdatesSubscription = self._pusher:isDoingWork()
		:filter( function( isDoingWork )
			return isDoingWork
		end )
		:subscribe( function()
			self._puller:checkOzForUpdates()
		end )
	
	self._catSpec = catSpec
	
	-- Setup custom asset sync handlers
	
	self._puller:addAssetTypeHandler( 'asset', WFCatIdleDownloader.addWaitingForProxyAnnotation )
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:close()

	check( 'WFCatSyncClientImp' )
	
	if self._checkOzForUpdatesSubscription then
	
		self._checkOzForUpdatesSubscription:dispose()
		self._checkOzForUpdatesSubscription = nil
	
	end
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:startSync()

	check( 'WFCatSyncClientImp' )
	self._puller:getUpdatesContinuously()
	self._pusher:sendUpdatesContinuously()
	
	return Observable.empty()
	
end

------------------------------------------------------------------------

function syncClientImpMethods:stopSync()

	check( 'WFCatSyncClientImp' )
	self._puller:stopContinuousUpdates()
	self._pusher:stopContinuousUpdates()
	
	return Observable.empty()
	
end

------------------------------------------------------------------------

function syncClientImpMethods:startSyncInResidualMode()

	check( 'WFCatSyncClientImp' )
	
	self._puller:getUpdatesContinuously()

	return Observable.empty()
	
end

------------------------------------------------------------------------

function syncClientImpMethods:stopSyncInResidualMode()
	check( 'WFCatSyncClientImp' )
	
	self._puller:stopContinuousUpdates()

	return Observable.empty()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:pausePusherReturningCounter()

	check( 'WFCatSyncClientImp' )

	return self._pusher:pauseReturningCounter()

end

------------------------------------------------------------------------

function syncClientImpMethods:resumePusher()

	check( 'WFCatSyncClientImp' )

	self._pusher:resume()

	return Observable.empty()

end

--------------------------------------------------------------------------------

-- For Lr CC debug panel.

function syncClientImpMethods:_DEBUG_startMetadataPusher()

	check( 'WFCatSyncClientImp' )
	
	self._pusher:sendUpdatesContinuously()

	return Observable.empty()
	
end

------------------------------------------------------------------------

-- For Lr CC debug panel.

function syncClientImpMethods:_DEBUG_stopMetadataPusher()
	check( 'WFCatSyncClientImp' )
	
	self._pusher:stopContinuousUpdates()

	return Observable.empty()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:setUserIsOverQuota( ... )
	check( 'WFCatSyncClientImp' )
	
	self._pusher:setUserIsOverQuota( ... )
	
	return Observable.empty()
end

--------------------------------------------------------------------------------

function syncClientImpMethods:setUserIsExpired( ... )
	check( 'WFCatSyncClientImp' )
	
	self._pusher:setUserIsExpired( ... )
	
	return Observable.empty()
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isUploadingMetadata()

	check( 'WFCatSyncClientImp' )
	return self._pusher:isDoingWork():distinctUntilChanged()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isDownloadingMetadata()

	check( 'WFCatSyncClientImp' )
	return self._puller:isDoingWork()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getLatestDownloadedAssetIdsStream()

	check( 'WFCatSyncClientImp' )
	return self._puller:getLatestDownloadedAssetIdsStream()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getDeletedDocumentIdStream()

	check( 'WFCatSyncClientImp' )
	return self._puller:getDeletedDocumentIdStream()
		
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isOzCatalogDeleted()

	check( 'WFCatSyncClientImp' )
	return self._puller:isOzCatalogDeleted()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getRequiredUpdateStream()

	check( 'WFCatSyncClientImp' )
	return self._user:getRequiredUpdateStream()
	
end

--------------------------------------------------------------------------------

return Rx.Class( 'WFCatSyncClientImp', syncClientImpMethods )
