--[[----------------------------------------------------------------------------

WFCatTrash.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.logger
local WFLogger = require 'WFLogger'
local logger = WFLogger( 'WFCatTrash' )

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFConfiguration = require 'WFConfiguration'
local WFRxThrottledBatchedBulkDocValue = require "WFRxThrottledBatchedBulkDocValue"
local WFRxThrottledBatchedBulkApplyOverlay = require 'WFRxThrottledBatchedBulkApplyOverlay'
local WFPlatformInfo = require 'WFPlatformInfo'
local WFCapabilities = require 'WFCapabilities'

	-- WF.oz_client.managed_catalog
local WFCatIdUtils = require 'WFCatIdUtils'
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatTrashUtils = require 'WFCatTrashUtils'

--==============================================================================

local ENABLE_CLOUD_TRASH = false
if WFPlatformInfo.isDesktop then
	ENABLE_CLOUD_TRASH = WFConfiguration.Shared.enableCloudTrash
else
	ENABLE_CLOUD_TRASH = WFCapabilities.getCapability( "WFCatDelete.enableCloudTrash" ) and true or false
end

local WFCatTrash = {}

local underscoreByte = string.byte( '_' )

-- TODO [kcarlson 2019-02-19]: Replace calls trashLog with logger.infof
-- when we remove the ENABLE_CLOUD_TRASH config flag.
local trashLogf = ENABLE_CLOUD_TRASH and logger.warnf or logger.infof

local appendArray = KSTableUtils.appendArray
local deepCopy = KSTableUtils.deepCopy
local derefOpt = KSTableUtils.derefOpt
local setNested = KSTableUtils.setNested

local check = assert( WFArguments.check )

local WFCatIndexUtils_isAssetOnOz = assert( WFCatIndexUtils.isAssetOnOz )
local WFCatIndexUtils_isAssetRevisionOnOz = assert( WFCatIndexUtils.isAssetRevisionOnOz )

local createFromInfo = assert( WFCatTrashUtils.createFromInfo )
local getCreateWithRevisionId = assert( WFCatTrashUtils.getCreateWithRevisionId )
local originalExistsOnOz = assert( WFCatTrashUtils.originalExistsOnOz )
local getCreateFromId = assert( WFCatTrashUtils.getCreateFromId )

--------------------------------------------------------------------------------

local function isAssetOnOzOrWasPushed( docValue, pusherChangeCounter )

	local docValueChangeCounter = docValue._changeCounter or 0

	return WFCatIndexUtils_isAssetOnOz( docValue ) or docValueChangeCounter < pusherChangeCounter

end

--------------------------------------------------------------------------------

function WFCatTrash.overlayForRemoveTrashInfo( docValue )
	return {
		_id = docValue._id,
		_localOnly = {
			universal = {
				trash = { _delete = true }
			}
		}
	}
end

--------------------------------------------------------------------------------

function WFCatTrash.overlayForDeleteTrashDocs( docIds )

	local overlays = {}

	for i, docId in ipairs( docIds ) do

		overlays[ i ] = {
			_id = docId,
			_deleted = true,
			_localOnly = {
				universal = {
					trash = {
						isDeleteTrash = true, -- aka manual purge
					}
				}
			}
		}

	end

	return overlays

end

--------------------------------------------------------------------------------

local function makeDeleteTrashOverlaysAndFileDeletionList( catalog, batchOfTrashDocValuesToDelete, pusherChangeCounter, optAddFilesToDeleteFromDocValue )

	local trashAssetOverlays = {}
	local localFiles = {}

	for i, docValue in ipairs( batchOfTrashDocValuesToDelete ) do

		local trashAssetOverlay =  {
			_id = docValue._id,
			_deleted = true,
			_localOnly = {
				universal = {
					trash = {
						isDeleteTrash = true, -- aka manual purge
					}
				}
			}
		}

		if isAssetOnOzOrWasPushed( docValue, pusherChangeCounter ) and createFromInfo( docValue ) then
			setNested( trashAssetOverlay, { _delete = true }, '_localOnly', 'universal', 'trash', 'createFrom' )
		end

		trashAssetOverlays[ i ] = trashAssetOverlay

		if optAddFilesToDeleteFromDocValue then
			optAddFilesToDeleteFromDocValue( localFiles, docValue )
		end

	end

	return trashAssetOverlays, localFiles

end

--------------------------------------------------------------------------------

local function deleteDocOverlaysFromQueryResult( rows )

	local overlays = {}

	-- Oz will take care of deleting the associated documents so apply
	-- the '_dontsendtooz' label, these overlays will be grouped with
	-- other overlays that don't have this set so apply it to each entry.

	for i, row in ipairs( rows ) do

		overlays[ i ] = {
			_id = row[ 1 ],
			_deleted = true,
			_newLabel = '_dontsendtooz',
		}

	end

	return overlays

end

--------------------------------------------------------------------------------

local function updateDocOverlaysFromQueryResult( docStore, rows, assetIdToNewAssetIdMap )

	local overlays = {}

	for i, row in ipairs( rows ) do

		local docId = row[ 1 ]
		local assetId = row[ 2 ]

		local newAssetId = assetIdToNewAssetIdMap[ assetId ]

		if newAssetId then

			overlays[ #overlays + 1 ] = {
				_id = docId,
				_preserveLabels = true,
				asset = {
					id = newAssetId
				}
			}

		else
			logger:errorf( "Trash assetId doesn't exist for assetId %s", assetId )
		end

	end

	return overlays

end

--------------------------------------------------------------------------------

local function updateAlbumAssetIdOverlays( catalog, assetIds, assetIdToNewAssetIdMap )

	-- Note this function is used to update the asset ids in album assets,
	-- it is used for trash assets to handle the case where a user could be
	-- offline and we temporarily save off the album assets. Oz will delete these
	-- once it sends down new album asset ids that it generates.
	-- And this is also used for when assets are restored, in this case the
	-- restored album assets will just update the trash album_assets.

	local albumAssetTable = "album_asset"

	if WFPlatformInfo.isDesktop then
		albumAssetTable = "album_asset_v2"
	end

	return catalog:indexStore():queryOnce {
		-- get all the album_assets
		sql = [[SELECT aa.docId, aa.assetId FROM ]] .. albumAssetTable .. [[ AS aa JOIN TEMP_ids ON TEMP_ids.docId = aa.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return updateDocOverlaysFromQueryResult( catalog:docStore(), rows, assetIdToNewAssetIdMap )
	end )

end

--------------------------------------------------------------------------------

local function updateCommentAssetsIdOverlays( catalog, assetIds, assetIdToNewAssetIdMap )

	return catalog:indexStore():queryOnce {
		-- get all the comments related to this asset
		sql = [[SELECT comment.docId, comment.assetId FROM comment JOIN TEMP_ids ON TEMP_ids.docId = comment.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return updateDocOverlaysFromQueryResult( catalog:docStore(), rows, assetIdToNewAssetIdMap )
	end )

end

--------------------------------------------------------------------------------

local function updateFavoriteAssetIdOverlays( catalog, assetIds, assetIdToNewAssetIdMap )

	return catalog:indexStore():queryOnce {
		-- get all the favorites related to this asset
		sql = [[SELECT spacefavorite.docId, spacefavorite.assetId FROM spacefavorite JOIN TEMP_ids ON TEMP_ids.docId = spacefavorite.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return updateDocOverlaysFromQueryResult( catalog:docStore(), rows, assetIdToNewAssetIdMap )
	end )

end

--------------------------------------------------------------------------------

local function deleteAlbumAssetOverlays( catalog, assetIds )

	local albumAssetTable = "album_asset"

	if WFPlatformInfo.isDesktop then
		albumAssetTable = "album_asset_v2"
	end

	return catalog:indexStore():queryOnce {
		-- get all the album_assets
		sql = [[SELECT aa.docId FROM ]] .. albumAssetTable .. [[ AS aa JOIN TEMP_ids ON TEMP_ids.docId = aa.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return deleteDocOverlaysFromQueryResult( rows )
	end )

end

--------------------------------------------------------------------------------

local function deleteCommentAssetOverlays( catalog, assetIds )

	return catalog:indexStore():queryOnce {
		-- get all the comments
		sql = [[SELECT comment.docId FROM comment JOIN TEMP_ids ON TEMP_ids.docId = comment.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return deleteDocOverlaysFromQueryResult( rows )
	end )

end

--------------------------------------------------------------------------------

local function deleteFavoriteAssetOverlays( catalog, assetIds )

	return catalog:indexStore():queryOnce {
		-- get all the favorites related to this asset
		sql = [[SELECT spacefavorite.docId, spacefavorite.assetId FROM spacefavorite JOIN TEMP_ids ON TEMP_ids.docId = spacefavorite.assetId]],
		tempIds = assetIds,
		quiesceUntilUpToDate = true,
	}
	:map( function( rows )
		return deleteDocOverlaysFromQueryResult( rows )
	end )

end

--------------------------------------------------------------------------------

local function createCopyOfAssetDocValue( docValue )

	-- Note this copies the _localOnly section to the result to preserve
	-- the local files and other fields.

	local copyOfDocValue = deepCopy( docValue )

	for k, _ in pairs( copyOfDocValue ) do
		if string.byte( k, 1 ) == underscoreByte
			and k ~= '_localOnly' then
				copyOfDocValue[ k ] = nil
		end
	end

	copyOfDocValue._id = KSIdGenerator.uuidInLowercaseHex()

	return copyOfDocValue

end

--------------------------------------------------------------------------------

local function setCreateWithRevId( origDocValue, trashDocOrRestoredDoc, assetIsOnOz )

	-- Specify _newRevId on the docValue and save it createWithRevId below. This
	-- is done in case another revision is added to the asset before the
	-- restore is issued. This is done to make sure a subsequent update
	-- will be sent with the modifications.

	-- It is also needed for trash asset to handle the case where you
	-- you do a modify delete sync.

	-- This revision id will be applied to the newly created asset (trash or restored).

	-- Note: createWithRevId does not belong in the createFrom section because
	-- it refers to information about the asset that the new asset is created from.
	-- createWithRevId is about the asset being created.

	local createWithRevId = KSIdGenerator.uuidInLowercaseHex()
	trashDocOrRestoredDoc._newRevId = createWithRevId
	setNested( trashDocOrRestoredDoc, createWithRevId, '_localOnly', 'universal', 'trash', 'createWithRevId')

	-- To handle docs that have been modified and then subsequently deleted and
	-- restored while offline (possibly multiple times)
	-- Generate a new revId to indicate that we should split the document
	-- into two revisions in addOverlaysForPreviouslyModifiedDocValues.

	-- The first revision won't ever go to oz but it's id will be used as
	-- the initial restore or trash asset revision id and conflict resolution
	-- will cause the second revision to be sent as an update.

	if not assetIsOnOz then

		local origDocCreateWithRevId = getCreateWithRevisionId( origDocValue )
		if origDocCreateWithRevId and origDocValue._revId ~= origDocCreateWithRevId then
			trashDocOrRestoredDoc._newRevId = KSIdGenerator.uuidInLowercaseHex()
		end

	end

end

--------------------------------------------------------------------------------

local function setOrUpdateCreateFrom( docValue, docIsOnOz, docValueRevId, trashDocOrRestoredDoc, lastTrashUserCreated, lastTrashPurgeDate )

	local originalId = docValue._id

	-- If creating a trash or restored asset from an asset that isn't on oz use
	-- it's prior createFrom info so that this asset will be created from the
	-- originally deleted asset that existed on oz (if there was one).

	if docIsOnOz then

		-- This is information that is used for the asset that the trash
		-- or restored asset will be created from.

		local createFrom = {
			id = originalId,
			revision_id = docValueRevId or docValue._revId,
			subtype = docValue.subtype
		}

		setNested( trashDocOrRestoredDoc, createFrom, '_localOnly', 'universal', 'trash', 'createFrom' )

	else

		-- createFrom was already copied from the original docValue in
		-- createCopyOfAssetDocValue (if it had one).
		-- So no need to set it here and we don't want to clear it as we want
		-- the trash document to refer to the original asset on oz.
		-- This is done to eliminate intermediate offline delete restore pairs
		-- for pattern [delete restore]+ delete

		docIsOnOz = originalExistsOnOz( docValue )
		if docIsOnOz then
			originalId = getCreateFromId( docValue )			
		end

	end

	if docIsOnOz then

		-- keep track of the last asset that this was generated from in the event
		-- that we need to create an intermediate asset to satisfy oz's requirement
		-- that a restore must be from a trash asset, and a delete generating a trash
		-- asset must be from a non-trash asset.
		setNested( trashDocOrRestoredDoc, docValue._id, '_localOnly', 'universal', 'trash', 'createFrom', 'lastAssetId' )
		setNested( trashDocOrRestoredDoc, docValue._revId, '_localOnly', 'universal', 'trash', 'createFrom', 'lastRevisionId' )
		setNested( trashDocOrRestoredDoc, lastTrashUserCreated, '_localOnly', 'universal', 'trash', 'createFrom', 'lastTrashUserCreated' )
		setNested( trashDocOrRestoredDoc, lastTrashPurgeDate, '_localOnly', 'universal', 'trash', 'createFrom', 'lastTrashPurgeDate' )

		-- Copy the links over so we know whether we have already uploaded them, and in case
		-- we need to download from them if the user is out of storage or the account is expired.

		local binaryLinksToCopy = {
			'/rels/master',
			'/rels/proxy_type/720p',
			'/rels/proxy_type/panorama',
			'/rels/proxy_type/2560',
			'/rels/xmp/develop',
			'/rels/xmp',
		}

		for _, linkName in ipairs( binaryLinksToCopy ) do

			local linkToBinaryOnOz = derefOpt( docValue, '_serverProvidedMetadata', 'links', linkName )
			if linkToBinaryOnOz then
				setNested( trashDocOrRestoredDoc, linkToBinaryOnOz, '_localOnly', 'universal', 'trash', 'createFrom', 'links', linkName )
				setNested( trashDocOrRestoredDoc, true, '_localOnly', 'universal', 'trash', 'uploaded', linkName)
			end

		end

	end

	return originalId

end

--------------------------------------------------------------------------------

function WFCatTrash.createTrashAssetFromDocValue( docValue, docIsOnOz, originalNewRevId, timeOfDelete, purgeDate )

	local trashDoc = createCopyOfAssetDocValue( docValue )

	-- deleted doc should be same as original except for
	-- id, subtype, and userCreated.
	-- Use passed in timeOfDelete so deleted assets could be batched together.

	if not docValue.subtype:find( 'deleted_' ) then
		trashDoc.subtype = 'deleted_' .. docValue.subtype
	end
	trashDoc.userCreated = timeOfDelete
	trashDoc.purgeDate = purgeDate

	local userCreated = docValue.userCreated
		or derefOpt( docValue, 'importSource', 'importTimestamp' )
		or derefOpt( docValue, '_serverProvidedMetadata', 'created' )

	setCreateWithRevId( docValue, trashDoc, docIsOnOz )

	local originalId = setOrUpdateCreateFrom( docValue, docIsOnOz, originalNewRevId, trashDoc )

	trashDoc.original = {
		id = originalId,
		userCreated = userCreated,
	}

	-- For mobile, remove chuncked upload info, so trash could restart upload from the beginning
	if WFPlatformInfo.isAndroid then

		if derefOpt( docValue, '_localOnly', 'originalFilePath' ) ~= nil then
			setNested( trashDoc, nil, '_localOnly', 'chunkedUpload' )
			setNested( trashDoc, nil, '_localOnly', 'original', 'uploaded' )
		end

	elseif WFPlatformInfo.isiOS then

		if derefOpt( trashDoc, '_localOnly', 'files', 'master' ) ~= nil then
			setNested( trashDoc, nil, '_localOnly', 'files', 'master', 'chunkedUpload' )
			setNested( trashDoc, nil, '_localOnly', 'files', 'master', 'uploaded' )
		end

	end

	return trashDoc

end

local createTrashAssetFromDocValue = WFCatTrash.createTrashAssetFromDocValue

--------------------------------------------------------------------------------

local function createOverlaysForAssetsMovedToTrash( catalog, pusherChangeCounter, timeOfDelete, purgeDate, batchOfDocValuesToDelete )

	-- Take a buffer of original docValues and create overlays to create
	-- trash docValues which contain an entire copy of most of the original asset,
	-- an overlay for the tombstones of the original assets being deleted,
	-- and an overlay for any modifications that were made to the asset
	-- being deleted that were not yet synced to oz.

	-- Trash assets and their corresponding delete of the original asset are
	-- interleved together to make the operation more atomic.  If users were
	-- to kill the application both operations would exist next to each other
	-- and most likely be executed together, assuming the app didn't die in
	-- between the two while the docstore was updating them.

	-- When we issued the operations separately it was possible for both the
	-- trash and original asset to exist after the app died resulting in
	-- many duplicated assets when batch sizes were greater than 100.

	local modifiedAssetIds = {}
	local positionOfModifiedAssetForId = {}

	for i, docValue in ipairs( batchOfDocValuesToDelete ) do

		local originalExistsOnOz = docValue._serverProvidedMetadata ~= nil

		if originalExistsOnOz and not WFCatIndexUtils_isAssetRevisionOnOz( docValue, docValue._revId ) then

			-- We need to make two revisions of modified documents that exist on oz
			-- as the delete operation in WFCatMetadataPusher isn't uploading any
			-- document changes in the payload (it just refers to the asset and revId
			-- for both deleted and trash asset.

			-- So make a list of assets that need updates and use their 'oz' doc value
			-- as the starting value for the original trash asset update.

			local nextModifiedAssetId = #modifiedAssetIds + 1
			modifiedAssetIds[ nextModifiedAssetId ] = docValue._id

			-- Since all assets are not modified make a look up table we will
			-- check against later when we make the trash asset.
			positionOfModifiedAssetForId[ docValue._id ] = nextModifiedAssetId
		end

	end

	-- Get the oz version of assets that were modified before they were deleted
	-- and use that as the starting doc value instead of the current.
	modifiedAssetIds.label = 'oz'
	local modifiedDocValues = #modifiedAssetIds > 0 and
		catalog:docStore():bulkDocValue( modifiedAssetIds ) or
			Observable.just( {} )

	return modifiedDocValues:map( function( batchOfOzDocValues )

		local assetOverlays = {}
		local assetIdToNewAssetIdMap = {}

		for i, docValue in ipairs( batchOfDocValuesToDelete ) do

			local originalNewRevId = KSIdGenerator.uuidInLowercaseHex() -- revId of tombstone

			local docIsOnOz = isAssetOnOzOrWasPushed( docValue, pusherChangeCounter )

			local ozDocValue = batchOfOzDocValues[ positionOfModifiedAssetForId[ docValue._id ] ]

			local trashDocValue
			local modifiedTrashDocValue = nil

			if ozDocValue then

				trashDocValue = createTrashAssetFromDocValue( ozDocValue, docIsOnOz, originalNewRevId, timeOfDelete, purgeDate )
				modifiedTrashDocValue = createTrashAssetFromDocValue( docValue, docIsOnOz, originalNewRevId, timeOfDelete, purgeDate )

				-- createTrashAssetFromDocValue generates a new id, keep the trashDocValue id.
				modifiedTrashDocValue._id = trashDocValue._id

				-- The modification to the trash doc value doesn't need to update any
				-- of the trash info and we want to preserve the createWithRevId so the
				-- modification will cause a conflict generating a subsequent update.
				setNested( modifiedTrashDocValue, nil, '_localOnly', 'universal', 'trash' )

			else

				trashDocValue = createTrashAssetFromDocValue( docValue, docIsOnOz, originalNewRevId, timeOfDelete, purgeDate )

			end

			-- Map the original asset id to the new trash asset id so we
			-- know how to update the associated docs (album_assets, comments, favorites).
			assetIdToNewAssetIdMap[ docValue._id ] = trashDocValue._id

			-- Create the trash asset.
			assetOverlays[ #assetOverlays + 1 ] = trashDocValue

			-- Delete the original asset, however don't send the delete's to oz, the trash
			-- assets create will be swapped out with a delete of the original asset that
			-- they refer to. This is done to eliminate the need of sending intermediate
			-- delete/restore/delete's when the user repeats the sequence while offline.
			assetOverlays[ #assetOverlays + 1 ] = {
				_id = docValue._id,
				_newRevId = originalNewRevId,
				_newLabel = '_dontsendtooz',
				_deleted = true,
				_localOnly = {
					deletedLocally = true, -- signals that file deletions are handled
					universal = {
						trash = {
							trash_asset = {
								id = trashDocValue._id,
								timeOfDelete = timeOfDelete,
								revision_id = trashDocValue._newRevId
							}
						}
					}
				}
			}

			-- If there are any, apply modifications that took place to the original
			-- asset that happened after it was last pushed up before it was deleted.
			if modifiedTrashDocValue then
				assetOverlays[ #assetOverlays + 1 ] = modifiedTrashDocValue
			end

		end

		return assetOverlays, assetIdToNewAssetIdMap

	end )

end

--------------------------------------------------------------------------------

function WFCatTrash.createRestoredAssetFromDocValue( trashDocValue, trashDocIsOnOz, timeOfRestore )

	local restoredDoc = createCopyOfAssetDocValue( trashDocValue )

	restoredDoc.subtype = string.match( trashDocValue.subtype, '^deleted_(.*)' )
	restoredDoc.userCreated = derefOpt( trashDocValue, 'original', 'userCreated' )

	if restoredDoc.subtype == nil then
		logger:errorf( "Restored Document subtype is nil, trashDocValue subtype was %s", trashDocValue.subtype )
	end

	-- Use passed in timeOfRestore so assets can be batched together properly.
	restoredDoc.userUpdated = timeOfRestore
	restoredDoc.original = nil
	restoredDoc.purgeDate = nil

	setCreateWithRevId( trashDocValue, restoredDoc, trashDocIsOnOz )

	setOrUpdateCreateFrom( trashDocValue, trashDocIsOnOz, trashDocValue._revId, restoredDoc, trashDocValue.userCreated, trashDocValue.purgeDate )

	-- For mobile, remove chuncked upload info, so trash could restart upload from the beginning
	if WFPlatformInfo.isAndroid then

		if derefOpt( restoredDoc, '_localOnly', 'originalFilePath' ) ~= nil then
			setNested( restoredDoc, nil, '_localOnly', 'chunkedUpload' )
			setNested( restoredDoc, nil, '_localOnly', 'original', 'uploaded' )
		end

	elseif WFPlatformInfo.isiOS then

		if derefOpt( restoredDoc, '_localOnly', 'files', 'master' ) ~= nil then
			setNested( restoredDoc, nil, '_localOnly', 'files', 'master', 'chunkedUpload' )
			setNested( restoredDoc, nil, '_localOnly', 'files', 'master', 'uploaded' )
		end

    end

	return restoredDoc

end

local createRestoredAssetFromDocValue = WFCatTrash.createRestoredAssetFromDocValue

--------------------------------------------------------------------------------

local function createOverlaysForRestoredTrashAssets( docStore, pusherChangeCounter, timeOfRestore, batchOfTrashDocValues )

	-- Given a batch of trash doc values to restore, generate overlays for
	-- the creation of the restored asset, deletion of the trash asset, and
	-- a subsequent update to the restored asset if changes were made prior
	-- to the asset's deletion.

	local batchOfOverlays = {}
	local assetIdToNewAssetIdMap = {}

	for i, trashDocValue in ipairs( batchOfTrashDocValues ) do

		local trashDocIsOnOz = isAssetOnOzOrWasPushed( trashDocValue, pusherChangeCounter )
		local restoredDoc = createRestoredAssetFromDocValue( trashDocValue, trashDocIsOnOz, timeOfRestore )

		local restoredDocModifyOverlay = nil
		local createWithRevId = getCreateWithRevisionId( restoredDoc )
		if restoredDoc._newRevId ~= createWithRevId then

			-- Make sure the restoredDoc uses createWithRevId it was changed only
			-- to indicate a prior change was applied and we need to generate
			-- a modification.
			restoredDoc._newRevId = createWithRevId

			-- Asset was modified before it was originally deleted and that modification
			-- was not synced up, create an extra overlay so that the restore or
			-- delete operation will result in a conflict.

			-- Since we don't know what fields were modified we will
			-- have the modified doc be a copy of the restored doc without
			-- the _localOnly section. The restoredDoc will have the
			-- _localOnly section, but any of the user changable content will be
			-- removed from it.  The 3-way merge in the conflict resolver will end up
			-- sending just what was changed anyways.
			restoredDocModifyOverlay = deepCopy( restoredDoc )

			-- Modification should create a new revId we don't need to know what it
			-- is so clear out _newRevId which was copied from the restored doc.
			restoredDocModifyOverlay._newRevId = nil

			-- The modification to the restored doc value doesn't need to update any
			-- of the trash info and we want to preserve the createWithRevId so the
			-- modification will cause a conflict generating a subsequent update.
			setNested( restoredDocModifyOverlay, nil, '_localOnly', 'universal', 'trash' )

			-- restoredDocModifyOverlay will apply all these fields in it's overlay,
			-- clear them out on the restoredDoc's first revision so we will create a conflict
			-- with what oz has and the 3 way merge will sort out only what actually
			-- changed.

			-- Remove any fields from the restoredDoc that the user can change
			-- develop settings, Stars, Flags, Title, Caption, Copyright
			-- Capture Date, location, city, state, country, keywords.
			restoredDoc.develop = nil
			restoredDoc.xmp = nil -- location, city, state, country, captureDate, keywords, Title, Caption, Copyright
			restoredDoc.ratings = nil -- stars
			restoredDoc.reviews = nil -- pick/reject flags
			restoredDoc.captureDate = nil -- if nil presents an issue, change it to be a fraction of a second earlier

			-- make it so the new userModified date is sent up as well.
			restoredDoc.userModified = nil

		end

		-- Map the trash asset id to the new restored asset id so we
		-- know how to update the associated docs (album_assets, comments, favorites).
		assetIdToNewAssetIdMap[ trashDocValue._id ] = restoredDoc._id

		batchOfOverlays[ #batchOfOverlays + 1 ] = restoredDoc

		-- don't need to send the actual delete to oz as oz will do that as part
		-- of the restore.
		batchOfOverlays[ #batchOfOverlays + 1 ] = {
			_id = trashDocValue._id,
			_deleted = true,
			_newLabel = '_dontsendtooz',
			_localOnly = {
				deletedLocally = true, -- signals that file deletions are handled
				universal = {
					trash = {
						restoredTo = {
							id = restoredDoc._id,
							timeOfRestore = timeOfRestore,
						}
					}
				}
			}
		}

		if restoredDocModifyOverlay then
			batchOfOverlays[ #batchOfOverlays + 1 ] = restoredDocModifyOverlay
		end

	end

	return batchOfOverlays, assetIdToNewAssetIdMap

end

--------------------------------------------------------------------------------

local function getAssetUuidsFromImageIds( catalog, imageIds, assetIdFromLocalIdFn )

	-- return Observable which fires only once with the list of oz-asset-ids
	-- converting local ids to oz-asset-ids if needed.

	-- Also ensures imageIds are either all local or all oz-asset-ids.

	local assetUuids

	if type( imageIds[ 1 ] ) == 'number' then

		for _, imageId in ipairs( imageIds ) do
			assert( type( imageId ) == 'number', 'All asset ids must be either local or global ids but not a mix' )
		end

		assetUuids = assetIdFromLocalIdFn( catalog, imageIds )

	elseif type( imageIds[ 1 ] ) == 'string' then

		for _, imageId in ipairs( imageIds ) do
			assert( type( imageId ) == 'string', 'All asset ids must be either local or global ids but not a mix' )
		end

		assetUuids = Observable.just( imageIds )

	else

		assert( false, 'Asset ids must all be uuid strings, or all be non-zero positive integers' )

	end

	return assetUuids

end

--------------------------------------------------------------------------------

function WFCatTrash.getAssetUuidsFromTrashImageIds( catalog, imageIds )

	-- Convert a local trashAsset localId to a UUID.  If imageIds
	-- contain UUIDs already do nothing.

	-- Note that his will only work for trash asset local Ids.
	-- Regular asset Ids should use WFCatIdUtils.getOzAssetIdsForLocalIds

	return getAssetUuidsFromImageIds( catalog, imageIds, WFCatIdUtils.getOzTrashAssetIdsForLocalIds )

end

--------------------------------------------------------------------------------

local function getAssetIdsFromBatchOfDocValues( batchOfDocValues )

	local batchOfIds = {}

	for i, docValue in ipairs( batchOfDocValues ) do
		batchOfIds[ i ] = docValue._id
	end

	return batchOfIds

end

--------------------------------------------------------------------------------

local pusherControlDisabled = {
	pauseReturningCounter = function()
		return Observable.just( 0 )
	end,
	resume = function()
	end
}

--------------------------------------------------------------------------------

function WFCatTrash.moveAssetsToTrash( catalog, imageIds, pusherControl, stackRelatedOverlays )

	-- This currently deletes stacks immediately so they aren't restored.

	check( 'table', 'table', '?table', '?function' )

	pusherControl = pusherControl or pusherControlDisabled

	stackRelatedOverlays = stackRelatedOverlays or function( catalog, assetIds )
		-- iOS and Android don't do anything with stacks and will return an empty table.
		return Observable.just( {} )
	end

	return Observable.combineLatest( function( pusherChangeCounter, purgeAssetsAfter, assetIds )

			trashLogf( logger, "moveAssetsToTrash (%d assets) assetIds = %s", #assetIds, assetIds )

			local timeOfDelete = KSDate.timeInIso8601Format()
			local purgeDate = KSDate.shiftIso8601DateTime( timeOfDelete, purgeAssetsAfter )

			return pusherChangeCounter, assetIds, timeOfDelete, purgeDate

		end,
		pusherControl.pauseReturningCounter(),
		catalog:ozUser():purgeAssetsAfterValue(),
		getAssetUuidsFromImageIds( catalog, imageIds, WFCatIdUtils.getOzAssetIdsForLocalIds )
	)

	:flatMap( function( pusherChangeCounter, assetIds, timeOfDelete, purgeDate )

		local docStore = catalog:docStore()

		-- Fully process the deletes in batches in case the app dies or gets killed.
		return WFRxThrottledBatchedBulkDocValue( docStore, assetIds, 100, 10 )

		:flatMap( function( batchOfDocValuesToDelete )

			local batchOfAssetIds = getAssetIdsFromBatchOfDocValues( batchOfDocValuesToDelete )

			return Observable.combineLatest(
				function( stackOverlays, assetOverlays, assetIdToTrashAssetIdMap )

					local stackOverlaysOz = stackOverlays[ 'oz' ] or {}
					local stackOverlaysLocal = stackOverlays[ 'local' ] or {}

					-- Combine the stackOverlays and the assetOverlays, making sure that the
					-- stackOverlays are first, then the assets...
					local combinedStackAndAssetOverlays = {}
					for i, stackOverlay in ipairs( stackOverlaysOz ) do
						stackOverlay._newLabel = '_dontsendtooz'
						combinedStackAndAssetOverlays[ i ] = stackOverlay
					end

					appendArray( combinedStackAndAssetOverlays, stackOverlaysLocal )

					appendArray( combinedStackAndAssetOverlays, assetOverlays )

					return batchOfAssetIds, combinedStackAndAssetOverlays, assetIdToTrashAssetIdMap

				end,
				stackRelatedOverlays( catalog, batchOfAssetIds ),
				createOverlaysForAssetsMovedToTrash( catalog, pusherChangeCounter,
					timeOfDelete, purgeDate, batchOfDocValuesToDelete ) )

		end )

		:flatMap( function( batchOfAssetIds, batchOfOverlays, assetIdToTrashAssetIdMap )

			-- Update all the associated documents (album_assets, comments, and favorites)
			-- to refer to the trash asset instead instead of the original asset.
			-- These all use the preserve 'oz' label enable to only send the associated
			-- documents that were created before the delete.

			return Observable.combineLatest(
				function( albumAssetOverlays, commentOverlays, favoriteOverlays )

					appendArray( batchOfOverlays, albumAssetOverlays )
					appendArray( batchOfOverlays, commentOverlays )
					appendArray( batchOfOverlays, favoriteOverlays )

					return batchOfOverlays

				end,
				updateAlbumAssetIdOverlays( catalog, batchOfAssetIds, assetIdToTrashAssetIdMap ),
				updateCommentAssetsIdOverlays( catalog, batchOfAssetIds, assetIdToTrashAssetIdMap ),
				updateFavoriteAssetIdOverlays( catalog, batchOfAssetIds, assetIdToTrashAssetIdMap )
			)

		end )

		:flatMap( function( batchOfOverlays )

			-- Apply all the stack overlays, trash asset creates, modifications
			-- and original asset deletes, and assoc doc updates in the same
			-- bulkApplyOverlay to make it more atomic. As bulkApplyOverlay
			-- has a transaction wrapped around it.

			return docStore:bulkApplyOverlay( batchOfOverlays )

		end )

	end )
	:takeLast( 0 )
	:tap( nil, pusherControl.resume, pusherControl.resume )

end

--------------------------------------------------------------------------------

function WFCatTrash.restoreTrashAssets( catalog, trashImageIds, pusherControl )

	check( 'table', 'table', '?table' )

	pusherControl = pusherControl or pusherControlDisabled

	local docStore = catalog:docStore()

	return Observable.combineLatest(
		pusherControl.pauseReturningCounter(),
		getAssetUuidsFromImageIds( catalog, trashImageIds, WFCatIdUtils.getOzTrashAssetIdsForLocalIds )
	)

	:flatMap( function( pusherChangeCounter, trashAssetIds )

		trashLogf( logger, "restoreTrashAssets (%d assets) assetIds = %s", #trashAssetIds, trashAssetIds )

		local timeOfRestore = KSDate.timeInIso8601Format()

		-- Fully process the restores in batches in case the app dies or gets killed.
		return WFRxThrottledBatchedBulkDocValue( docStore, trashAssetIds, 100, 10 )

		:map( function( batchOfDocValuesToRestore )

			local batchOfTrashAssetIds = getAssetIdsFromBatchOfDocValues( batchOfDocValuesToRestore )

			local batchOfOverlays, trashAssetIdToRestoredAssetIdMap =
				createOverlaysForRestoredTrashAssets( docStore, pusherChangeCounter, timeOfRestore, batchOfDocValuesToRestore )

			return batchOfTrashAssetIds, batchOfOverlays, trashAssetIdToRestoredAssetIdMap

		end )

		:flatMap( function( batchOfTrashAssetIds, batchOfOverlays, trashAssetIdToRestoredAssetIdMap )

			-- Generate overlays for all the associated documents (album_assets, comments, and favorites)
			-- to update them to refer to the restored asset instead instead of the trash asset.
			-- These all use the _preserveLabel flag on the overlay so we only send the associated
			-- documents that were newly created or deleted before the asset delete.

			return Observable.combineLatest(
				function( albumAssetOverlays, commentOverlays, favoriteOverlays )

					appendArray( batchOfOverlays, albumAssetOverlays )
					appendArray( batchOfOverlays, commentOverlays )
					appendArray( batchOfOverlays, favoriteOverlays )

					return batchOfOverlays

				end,
				updateAlbumAssetIdOverlays( catalog, batchOfTrashAssetIds, trashAssetIdToRestoredAssetIdMap ),
				updateCommentAssetsIdOverlays( catalog, batchOfTrashAssetIds, trashAssetIdToRestoredAssetIdMap ),
				updateFavoriteAssetIdOverlays( catalog, batchOfTrashAssetIds, trashAssetIdToRestoredAssetIdMap )
			)

		end )

		:flatMap( function( batchOfOverlays )

			-- Apply all restored asset creates, modifications
			-- and trash asset deletes, and assoc doc updates in the same
			-- bulkApplyOverlay to make it more atomic. As bulkApplyOverlay
			-- has a transaction wrapped around it.

			return docStore:bulkApplyOverlay( batchOfOverlays )

		end )

	end )
	:takeLast( 0 )
	:tap( nil, pusherControl.resume, pusherControl.resume )

end

--------------------------------------------------------------------------------

function WFCatTrash.deleteTrashAssets( catalog, imageIds, pusherControl, optAddFilesToDeleteFromDocValue, optFileDeletionHandler )

	check( 'table', 'table', '?table', '?function', '?function' )

	pusherControl = pusherControl or pusherControlDisabled

	local docStore = catalog:docStore()

	return Observable.combineLatest(
		pusherControl.pauseReturningCounter(),
		getAssetUuidsFromImageIds( catalog, imageIds, WFCatIdUtils.getOzTrashAssetIdsForLocalIds )
	)

	:flatMap( function( pusherChangeCounter, assetIds )

		trashLogf( logger, "deleteTrashAssets (%d assets) assetIds = %s", #assetIds, assetIds )

		-- Fully process the deleteTrash in batches in case the app dies or gets killed.
		return WFRxThrottledBatchedBulkDocValue( docStore, assetIds, 100, 10 )

		:map( function( batchOfTrashDocValuesToDelete )

			local batchOfTrashAssetIds = getAssetIdsFromBatchOfDocValues( batchOfTrashDocValuesToDelete )

			local batchOfTrashAssetOverlays, batchOfLocalFilesToDelete =
				makeDeleteTrashOverlaysAndFileDeletionList( catalog, batchOfTrashDocValuesToDelete, pusherChangeCounter, optAddFilesToDeleteFromDocValue )

			return batchOfTrashAssetIds, batchOfTrashAssetOverlays, batchOfLocalFilesToDelete

		end )

		:flatMap( function( batchOfTrashAssetIds, batchOfOverlays, batchOfLocalFilesToDelete )

			return Observable.combineLatest(
				function( albumAssetOverlays, commentOverlays, favoriteOverlays )

					appendArray( batchOfOverlays, albumAssetOverlays )
					appendArray( batchOfOverlays, commentOverlays )
					appendArray( batchOfOverlays, favoriteOverlays )

					return batchOfTrashAssetIds, batchOfOverlays, batchOfLocalFilesToDelete

				end,
				deleteAlbumAssetOverlays( catalog, batchOfTrashAssetIds ),
				deleteCommentAssetOverlays( catalog, batchOfTrashAssetIds ),
				deleteFavoriteAssetOverlays( catalog, batchOfTrashAssetIds )
			)

		end )

		-- Delete all the trash assets (subtype deleted_image, deleted_video) from the doc store
		-- along with all the associated documents. This is done in the same bulkApplyOverlay
		-- to make it atomic for the batch as there is a transaction wrapped around bulkApplyOverlay.

		:flatMap( function( batchOfTrashAssetIds, batchOfOverlays, batchOfLocalFilesToDelete )

			return docStore:bulkApplyOverlay( batchOfOverlays )
			:map( function()

				-- Delete the local files related to these assets
				if optFileDeletionHandler then
					optFileDeletionHandler( batchOfTrashAssetIds, batchOfLocalFilesToDelete )
				end

				return true

			end )

		end )

	end )

	:takeLast( 0 )
	:tap( nil, pusherControl.resume, pusherControl.resume )

end

--------------------------------------------------------------------------------

local sortOrderSubject = Rx.BehaviorSubject( 'ASC' )

function WFCatTrash.setTrashAssetListSortOrder( order )

	sortOrderSubject:onNext( order )

end

--------------------------------------------------------------------------------

local function getNextRefreshForPurgeDate( mostImmediatePurgeDate )

	local now = KSDate.timeInIso8601Format()
	local nextCheckDelay = KSDate.differenceInSeconds( now, mostImmediatePurgeDate )
	if nextCheckDelay < 1 then
		nextCheckDelay = 1
	end
	
	return nextCheckDelay, now

end

--------------------------------------------------------------------------------

local nextListCheckSubject

function WFCatTrash.trashAssetList( catalog, fieldsToReturn, getMostImmediatePurgeDateFunction, encoder )

	nextListCheckSubject = Rx.BehaviorSubject( 0 )

	return Observable.combineLatest( sortOrderSubject, nextListCheckSubject:flatMapLatest( function( refreshInSeconds )
			return Rx.Observable.timer( refreshInSeconds, WFRxScheduler.sharedScheduler() )
		end ))
		:flatMap( function( ascDesc )
		
			local currentTime = KSDate.timeInIso8601Format()

			local indexStore = catalog:indexStore()
			
			return indexStore:query { 
				sql = string.format( "SELECT %s FROM trashAssets WHERE purgeDate > ? ORDER BY purgeDate %s", 
					fieldsToReturn, ascDesc ),
				currentTime,
				encoder = encoder,
			}
			
			:map( function( rows )
				
				local mostImmediatePurgeDate = getMostImmediatePurgeDateFunction( rows, ascDesc )
				
				if mostImmediatePurgeDate then
					local nextRefreshInSeconds, now = getNextRefreshForPurgeDate( mostImmediatePurgeDate )
					
					logger:infof( "WFCatTrash.trashAssetList: next check in %s seconds (time now is %s)", nextRefreshInSeconds, now )
					nextListCheckSubject:onNext( nextRefreshInSeconds )
				end
				
				return rows

			end )

		end )

end

--------------------------------------------------------------------------------

local nextCountCheckSubject

function WFCatTrash.countTrashAssets( catalog )

	nextCountCheckSubject = Rx.BehaviorSubject( 0 )
	
	return Observable.combineLatest( nextCountCheckSubject:flatMapLatest( function( refreshInSeconds )
			return Rx.Observable.timer( refreshInSeconds, WFRxScheduler.sharedScheduler() )
		end ))
		:flatMap( function()
		
			local currentTime = KSDate.timeInIso8601Format()
			
			return catalog:indexStore():query { 
				sql = "SELECT COUNT(*), purgeDate FROM trashAssets WHERE purgeDate > ? ORDER BY purgeDate ASC LIMIT 1",
				currentTime,
			}
			:tap( function( rows )
	
				local mostImmediatePurgeDate = rows[ 1 ] and rows[ 1 ][ 2 ]
				
				if mostImmediatePurgeDate then
					local nextRefreshInSeconds, now = getNextRefreshForPurgeDate( mostImmediatePurgeDate )
					
					logger:infof( "WFCatTrash.countTrashAssets: next check in %s seconds (time now is %s)", nextRefreshInSeconds, now )
					nextCountCheckSubject:onNext( nextRefreshInSeconds )
				end
	
			end )
			
		end )
		
end

--------------------------------------------------------------------------------

local nextAutoPurgeCheckSubject

function WFCatTrash.autoPurge( catalog )

	nextAutoPurgeCheckSubject = Rx.BehaviorSubject( 0 )

	return Observable.combineLatest( nextAutoPurgeCheckSubject:flatMapLatest( function( refreshInSeconds )
			return Rx.Observable.timer( refreshInSeconds, WFRxScheduler.sharedScheduler() )
		end ))
		:flatMap( function()
		
			local currentTime = KSDate.timeInIso8601Format()
			
			return catalog:indexStore():query { 
				sql = "SELECT docId FROM trashAssets WHERE purgeDate <= ? ORDER BY purgeDate ASC",
				currentTime,
			}
			
			:map( function( rows )
				
				local autoPurgedAssetOverlays = {}

				for i, row in ipairs( rows or {} ) do
					
					local autoPurgedAssetId = row[1]
					
					autoPurgedAssetOverlays[i] = {
						_id = autoPurgedAssetId,
						_deleted = true,
					}
			
				end
				
				autoPurgedAssetOverlays.label = '_dontsendtooz'

				return #autoPurgedAssetOverlays > 0 and 
					WFRxThrottledBatchedBulkApplyOverlay( catalog:docStore(), autoPurgedAssetOverlays, 100, 10 ) or 
					Observable.just( {} )
					
			end )
			
			:map( function() 
				return currentTime
			end )
		
		end )
		
		:flatMap( function( currentTime )
		
			return catalog:indexStore():query( {
				sql = "SELECT purgeDate FROM trashAssets WHERE purgeDate > ? ORDER BY purgeDate ASC LIMIT 1",
				currentTime
			} )
			:tap( function( rows )
				
				if rows[ 1 ] then
					
					local mostImmediatePurgeDate = rows[ 1 ][ 1 ]
				
					if mostImmediatePurgeDate then
						local nextRefreshInSeconds, now = getNextRefreshForPurgeDate( mostImmediatePurgeDate )
						
						logger:infof( "WFCatTrash.autoPurge: next check in %s seconds (time now is %s)", nextRefreshInSeconds, now )
						nextAutoPurgeCheckSubject:onNext( nextRefreshInSeconds )
					end
					
				end

			end )

		end )

end

--==============================================================================

return WFCatTrash
